from __future__ import annotations

from typing import TYPE_CHECKING

from nbxmpp.modules.base import BaseModule
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import NodeProcessed
from nbxmpp.protocol import Presence
from nbxmpp.structs import Hat
from nbxmpp.structs import HatData
from nbxmpp.structs import PresenceProperties
from nbxmpp.structs import StanzaHandler

if TYPE_CHECKING:
    from nbxmpp.client import Client


class Hats(BaseModule):
    def __init__(self, client: Client) -> None:
        BaseModule.__init__(self, client)
        self._client = client
        self.handlers = [
            StanzaHandler(
                name="presence",
                callback=self._process_hats,
                ns=Namespace.HATS,
                priority=15,
            ),
            StanzaHandler(
                name="presence",
                callback=self._process_hats,
                ns=Namespace.HATS_LEGACY,
                priority=15,
            ),
        ]

    def _process_hats(
        self,
        _client: Client,
        stanza: Presence,
        properties: PresenceProperties,
    ) -> None:
        hats = stanza.getTag("hats", namespace=Namespace.HATS)
        if hats is None:
            hats = stanza.getTag("hats", namespace=Namespace.HATS_LEGACY)
            if hats is None:
                return

        hat_data = HatData()
        for hat in hats.getTags("hat"):
            uri = hat.getAttr("uri")
            title = hat.getAttr("title")
            if not uri or not title:
                self._log.warning("Invalid hat received")
                self._log.warning(stanza)
                raise NodeProcessed

            lang = hat.getAttr("xml:lang")
            hat_data.add_hat(Hat(uri, title), lang)

        properties.hats = hat_data



# Copyright (C) 2021 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from typing import TYPE_CHECKING

from nbxmpp.errors import MalformedStanzaError
from nbxmpp.errors import StanzaError
from nbxmpp.modules.base import BaseModule
from nbxmpp.modules.util import process_response
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Iq
from nbxmpp.protocol import JID
from nbxmpp.protocol import NodeProcessed
from nbxmpp.simplexml import Node
from nbxmpp.structs import IqProperties
from nbxmpp.structs import RosterData
from nbxmpp.structs import RosterItem
from nbxmpp.structs import RosterPush
from nbxmpp.structs import StanzaHandler
from nbxmpp.task import iq_request_task

if TYPE_CHECKING:
    from nbxmpp.client import Client


class Roster(BaseModule):
    def __init__(self, client: Client) -> None:
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = [
            StanzaHandler(
                name="iq",
                callback=self._process_roster_push,
                typ="set",
                priority=15,
                ns=Namespace.ROSTER,
            ),
        ]

    @iq_request_task
    def request_roster(self, version: str | None = None):
        _task = yield

        ver_support = self._client.features.has_roster_version()
        if not ver_support:
            version = None

        if ver_support and version is None:
            version = ""

        self._log.info("Roster versioning supported: %s", ver_support)

        response = yield _make_request(version, ver_support)
        if response.isError():
            raise StanzaError(response)

        query = response.getTag("query", namespace=Namespace.ROSTER)
        if query is None:
            if not ver_support:
                raise MalformedStanzaError("query node missing", response)
            yield RosterData(None, version)

        pushed_items, version = self._parse_push(response, ver_support)
        yield RosterData(pushed_items, version)

    def _process_roster_push(
        self, _client: Client, stanza: Iq, properties: IqProperties
    ) -> None:
        from_ = stanza.getFrom()
        if from_ is not None:
            if not self._client.get_bound_jid().bare == from_:
                self._log.warning("Malicious Roster Push from %s", from_)
                raise NodeProcessed

        ver_support = self._client.features.has_roster_version()
        pushed_items, version = self._parse_push(stanza, ver_support)
        if len(pushed_items) != 1:
            self._log.warning(
                "Roster push does not have exactly one item: %s", pushed_items
            )
            self._log.warning(stanza)
            raise NodeProcessed

        item = pushed_items[0]
        properties.roster = RosterPush(item, version)

        self._log.info("Roster Push, version: %s", properties.roster.version)
        self._log.info(item)

        self._ack_roster_push(stanza)

    def _ack_roster_push(self, stanza: Iq) -> None:
        iq = Iq(
            "result",
            to=stanza.getFrom(),
            frm=stanza.getTo(),
            attrs={"id": stanza.getID()},
        )
        self._client.send_stanza(iq)

    @iq_request_task
    def delete_item(self, jid: JID | str):
        _task = yield

        response = yield _make_delete(jid)
        yield process_response(response)

    @iq_request_task
    def set_item(self, jid: JID, name: str | None, groups: set[str] | None = None):
        _task = yield

        response = yield _make_set(jid, name, groups)
        yield process_response(response)

    def _parse_push(
        self, stanza: Iq, ver_support: bool
    ) -> tuple[list[RosterItem], str | None]:
        query = stanza.getTag("query", namespace=Namespace.ROSTER)

        version = None
        if ver_support:
            version = query.getAttr("ver")
            if version is None:
                # raise MalformedStanzaError('ver attribute missing', stanza)
                # Prosody sometimes does not send ver attribute with some
                # community modules
                self._log.warning("no version attribute found")

        pushed_items: list[RosterItem] = []
        for item in query.getTags("item"):
            try:
                roster_item = RosterItem.from_node(item)
            except Exception as e:
                self._log.warning("Invalid roster item: %s", e)
                self._log.warning(stanza)
                continue

            pushed_items.append(roster_item)

        return pushed_items, version


def _make_delete(jid: JID | str) -> Iq:
    return Iq(
        "set",
        Namespace.ROSTER,
        payload=[Node("item", {"jid": jid, "subscription": "remove"})],
    )


def _make_set(jid: JID, name: str | None, groups: set[str] | None = None) -> Iq:
    if groups is None:
        groups = set()

    infos: dict[str, JID | str] = {"jid": jid}
    if name:
        infos["name"] = name
    iq = Iq("set", Namespace.ROSTER)
    query = iq.setQuery()
    item = query.addChild("item", attrs=infos)
    for group in groups:
        item.addChild("group").setData(group)
    return iq


def _make_request(version: str | None, roster_ver_support: bool) -> Iq:
    iq = Iq("get", Namespace.ROSTER)
    if version is None:
        version = ""

    if roster_ver_support:
        iq.setTagAttr("query", "ver", version)
    return iq



# Copyright (C) 2018 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from typing import TYPE_CHECKING

from nbxmpp.errors import MalformedStanzaError
from nbxmpp.modules.base import BaseModule
from nbxmpp.modules.bookmarks.util import build_storage_node
from nbxmpp.modules.bookmarks.util import parse_bookmarks
from nbxmpp.modules.util import raise_if_error
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Message
from nbxmpp.protocol import NodeProcessed
from nbxmpp.structs import BookmarkData
from nbxmpp.structs import MessageProperties
from nbxmpp.structs import StanzaHandler
from nbxmpp.task import iq_request_task

BOOKMARK_OPTIONS = {
    "pubsub#persist_items": "true",
    "pubsub#access_model": "whitelist",
}

if TYPE_CHECKING:
    from nbxmpp.client import Client


class PEPBookmarks(BaseModule):

    _depends = {
        "publish": "PubSub",
        "request_items": "PubSub",
    }

    def __init__(self, client: Client) -> None:
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = [
            StanzaHandler(
                name="message",
                callback=self._process_pubsub_bookmarks,
                ns=Namespace.PUBSUB_EVENT,
                priority=16,
            ),
        ]

    def _process_pubsub_bookmarks(
        self, _client: Client, stanza: Message, properties: MessageProperties
    ) -> None:
        if not properties.is_pubsub_event:
            return

        if properties.pubsub_event.node != Namespace.BOOKMARKS:
            return

        item = properties.pubsub_event.item
        if item is None:
            # Retract, Deleted or Purged
            return

        try:
            bookmarks = parse_bookmarks(item, self._log)
        except MalformedStanzaError as error:
            self._log.warning(error)
            self._log.warning(stanza)
            raise NodeProcessed

        if not bookmarks:
            self._log.info("Bookmarks removed")
            return

        pubsub_event = properties.pubsub_event._replace(data=bookmarks)
        self._log.info("Received bookmarks from: %s", properties.jid)
        for bookmark in bookmarks:
            self._log.info(bookmark)

        properties.pubsub_event = pubsub_event

    @iq_request_task
    def request_bookmarks(self):
        _task = yield

        items = yield self.request_items(Namespace.BOOKMARKS, max_items=1)
        raise_if_error(items)

        if not items:
            yield []

        bookmarks = parse_bookmarks(items[0], self._log)
        for bookmark in bookmarks:
            self._log.info(bookmark)

        yield bookmarks

    @iq_request_task
    def store_bookmarks(self, bookmarks: list[BookmarkData]):
        _task = yield

        self._log.info("Store Bookmarks")

        self.publish(
            Namespace.BOOKMARKS,
            build_storage_node(bookmarks),
            id_="current",
            options=BOOKMARK_OPTIONS,
            force_node_options=True,
        )



from __future__ import annotations

from test.lib.util import StanzaHandlerTest

import nbxmpp
from nbxmpp.language import LanguageRange
from nbxmpp.structs import BodyData


class TestBody(StanzaHandlerTest):

    def test_body_language(self):
        text = "> Anna wrote:\n> Hi, how are you?\nGreat"
        text_de = "> Anna schrieb\n>Hallo, wie geht es dir?\nGut"

        xml = """
            <message to='anna@example.com' id='message-id2' type='groupchat'>
              <body>%s</body>
              <body xml:lang='en-us'>%s</body>
              <body xml:lang='de-DE'>%s</body>
            </message>
        """ % (
            text,
            text,
            text_de,
        )

        message = nbxmpp.Message(node=xml)

        body_data = BodyData(message, None, None)
        body = body_data.get([LanguageRange(tag="en")])
        self.assertEqual(body, text)

        body = body_data.get(None)
        self.assertEqual(body, text)

        body = body_data.get([LanguageRange(tag="de")])
        self.assertEqual(body, text_de)

        body = body_data.get([LanguageRange(tag="ar")])
        self.assertEqual(body, text)

        xml = """
            <message to='anna@example.com' id='message-id2' type='groupchat'>
              <body xml:lang='en-us'>en</body>
              <body xml:lang='de-DE'>de</body>
            </message>
        """

        message = nbxmpp.Message(node=xml)

        body_data = BodyData(message)
        body = body_data.get([LanguageRange(tag="en")])
        self.assertEqual(body, "en")

        # Language not in body, pick any language
        body_data = BodyData(message)
        body = body_data.get([LanguageRange(tag="ar")])
        self.assertEqual(body, "de")

        # Supply range with priority
        body_data = BodyData(message)
        body = body_data.get([LanguageRange(tag="de"), LanguageRange(tag="en")])
        self.assertEqual(body, "de")

        body_data = BodyData(message)
        body = body_data.get([LanguageRange(tag="ar"), LanguageRange(tag="en")])
        self.assertEqual(body, "en")



# Copyright (C) 2003-2005 Alexey "Snake" Nezhdanov
#
# SPDX-License-Identifier: GPL-2.0-or-later

"""
Provides PlugIn class functionality to develop extentions for xmpppy
"""

import logging

log = logging.getLogger("nbxmpp.plugin")


class PlugIn:
    """
    Abstract xmpppy plugin infrastructure code, providing plugging in/out and
    debugging functionality

    Inherit to develop pluggable objects. No code change on the owner class
    required (the object where we plug into)

    For every instance of PlugIn class the 'owner' is the class in what the plug
    was plugged.
    """

    def __init__(self):
        self._exported_methods = []

    def PlugIn(self, owner, *args, **kwargs):
        """
        Attach to owner and register ourself and our _exported_methods in it.
        If defined by a subclass, call self.plugin(owner) to execute hook
        code after plugging
        """
        self._owner = owner
        log.info("Plugging %s __INTO__ %s", self, self._owner)
        if self.__class__.__name__ in owner.__dict__:
            log.debug("Plugging ignored: another instance already plugged.")
            return None
        self._old_owners_methods = []
        for method in self._exported_methods:
            if method.__name__ in owner.__dict__:
                self._old_owners_methods.append(owner.__dict__[method.__name__])
            owner.__dict__[method.__name__] = method
        if self.__class__.__name__.endswith("Dispatcher"):
            # FIXME: I need BOSHDispatcher or XMPPDispatcher on .Dispatcher
            # there must be a better way..
            owner.__dict__["Dispatcher"] = self
        else:
            owner.__dict__[self.__class__.__name__] = self

        # Execute hook
        if hasattr(self, "plugin"):
            return self.plugin(owner, *args, **kwargs)
        return None

    def PlugOut(self, *args, **kwargs):
        """
        Unregister our _exported_methods from owner and detach from it.
        If defined by a subclass, call self.plugout() after unplugging to
        execute hook code
        """
        log.info("Plugging %s __OUT__ of %s.", self, self._owner)
        for method in self._exported_methods:
            del self._owner.__dict__[method.__name__]
        for method in self._old_owners_methods:
            self._owner.__dict__[method.__name__] = method
        # FIXME: Dispatcher workaround
        if self.__class__.__name__.endswith("Dispatcher"):
            del self._owner.__dict__["Dispatcher"]
        else:
            del self._owner.__dict__[self.__class__.__name__]
        # Execute hook
        if hasattr(self, "plugout"):
            return self.plugout(*args, **kwargs)
        del self._owner
        return None

    @classmethod
    def get_instance(cls, *args, **kwargs):
        """
        Factory Method for object creation

        Use this instead of directly initializing the class in order to make
        unit testing easier. For testing, this method can be patched to inject
        mock objects.
        """
        return cls(*args, **kwargs)



# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

# XEP-0461: Message Replies

from __future__ import annotations

from typing import TYPE_CHECKING

from nbxmpp.modules.base import BaseModule
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import JID
from nbxmpp.protocol import Message
from nbxmpp.structs import MessageProperties
from nbxmpp.structs import ReplyData
from nbxmpp.structs import StanzaHandler

if TYPE_CHECKING:
    from nbxmpp.client import Client


class Replies(BaseModule):
    def __init__(self, client: Client) -> None:
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = [
            StanzaHandler(
                name="message",
                callback=self._process_message,
                ns=Namespace.REPLY,
                priority=15,
            )
        ]

    def _process_message(
        self, _client: Client, stanza: Message, properties: MessageProperties
    ) -> None:

        reply = stanza.getTag("reply", namespace=Namespace.REPLY)
        if reply is None:
            return

        to = reply.getAttr("to")
        if to is None:
            self._log.warning('Received reply without "to" attribute')
            return

        try:
            reply_to = JID.from_string(to)
        except Exception:
            self._log.warning("Invalid jid on reply element: %s", to)
            return

        reply_to_id = reply.getAttr("id")
        if reply_to_id is None:
            self._log.warning('Received reply without "id"')
            return

        properties.reply_data = ReplyData(to=reply_to, id=reply_to_id)



# Copyright (C) 2018 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from typing import TYPE_CHECKING

from nbxmpp.modules.base import BaseModule
from nbxmpp.modules.bits_of_binary import parse_bob_data
from nbxmpp.modules.dataforms import extend_form
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Message
from nbxmpp.structs import CaptchaData
from nbxmpp.structs import MessageProperties
from nbxmpp.structs import StanzaHandler

if TYPE_CHECKING:
    from nbxmpp.client import Client


class Captcha(BaseModule):
    def __init__(self, client: Client) -> None:
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = [
            StanzaHandler(
                name="message",
                callback=self._process_captcha,
                ns=Namespace.CAPTCHA,
                priority=40,
            ),
        ]

    def _process_captcha(
        self, _client: Client, stanza: Message, properties: MessageProperties
    ) -> None:
        captcha = stanza.getTag("captcha", namespace=Namespace.CAPTCHA)
        if captcha is None:
            return

        data_form = captcha.getTag("x", namespace=Namespace.DATA)
        if data_form is None:
            self._log.warning("Invalid captcha form")
            self._log.warning(stanza)
            return

        form = extend_form(node=data_form)
        bob_data = parse_bob_data(stanza)

        properties.captcha = CaptchaData(form=form, bob_data=bob_data)



# Copyright (C) 2020 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from typing import TYPE_CHECKING

from nbxmpp.errors import ChangePasswordStanzaError
from nbxmpp.errors import RegisterStanzaError
from nbxmpp.errors import StanzaError
from nbxmpp.modules.base import BaseModule
from nbxmpp.modules.register.util import _make_password_change_request
from nbxmpp.modules.register.util import _make_password_change_with_form
from nbxmpp.modules.register.util import _make_register_form
from nbxmpp.modules.register.util import _make_unregister_request
from nbxmpp.modules.register.util import _parse_register_data
from nbxmpp.modules.util import process_response
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Iq
from nbxmpp.protocol import JID
from nbxmpp.task import iq_request_task
from nbxmpp.util import get_form

if TYPE_CHECKING:
    from nbxmpp.client import Client


class Register(BaseModule):
    def __init__(self, client: Client) -> None:
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = []

    @iq_request_task
    def unregister(self, jid: JID | None = None):
        _task = yield

        response = yield _make_unregister_request(jid)
        yield process_response(response)

    @iq_request_task
    def request_register_form(self, jid: JID | None = None):
        _task = yield

        if jid is None:
            jid = self._client.domain

        response = yield Iq("get", Namespace.REGISTER, to=jid)
        if response.isError():
            raise StanzaError(response)

        yield _parse_register_data(response)

    @iq_request_task
    def submit_register_form(self, form, jid: JID | None = None):
        _task = yield

        if jid is None:
            jid = self._client.domain

        response = yield _make_register_form(jid, form)
        if not response.isError():
            yield process_response(response)

        else:
            data = _parse_register_data(response)
            raise RegisterStanzaError(response, data)

    @iq_request_task
    def change_password(self, password: str):
        _task = yield

        response = yield _make_password_change_request(
            self._client.domain, self._client.username, password
        )
        if not response.isError():
            yield process_response(response)

        else:
            query = response.getQuery()
            if query is None:
                raise StanzaError(response)

            form = get_form(query, "jabber:iq:register:changepassword")
            if form is None or response.getType() != "modify":
                raise StanzaError(response)

            raise ChangePasswordStanzaError(response, form)

    @iq_request_task
    def change_password_with_form(self, form):
        _task = yield

        response = yield _make_password_change_with_form(self._client.domain, form)
        yield process_response(response)



# Copyright (C) 2018 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import hashlib
import logging

from nbxmpp.namespaces import Namespace
from nbxmpp.simplexml import Node
from nbxmpp.structs import BobData
from nbxmpp.util import b64decode

log = logging.getLogger("nbxmpp.m.bob")


def parse_bob_data(stanza: Node) -> BobData | None:
    data_node = stanza.getTag("data", namespace=Namespace.BOB)
    if data_node is None:
        return None

    cid = data_node.getAttr("cid")
    type_ = data_node.getAttr("type")
    max_age = data_node.getAttr("max-age")
    if max_age is not None:
        try:
            max_age = int(max_age)
        except Exception:
            log.exception(stanza)
            return None

    assert max_age is not None

    if cid is None or type_ is None:
        log.warning("Invalid data node (no cid or type attr): %s", stanza)
        return None

    try:
        algo_hash = cid.split("@")[0]
        algo, hash_ = algo_hash.split("+")
    except Exception:
        log.exception("Invalid cid: %s", stanza)
        return None

    bob_data = data_node.getData()
    if not bob_data:
        log.warning("No bob data found: %s", stanza)
        return None

    try:
        bob_data = b64decode(bob_data)
    except Exception:
        log.warning("Unable to decode data")
        log.exception(stanza)
        return None

    try:
        sha = hashlib.new(algo)
    except ValueError as error:
        log.warning(stanza)
        log.warning(error)
        return None

    sha.update(bob_data)
    if sha.hexdigest() != hash_:
        log.warning("Invalid hash: %s", stanza)
        return None

    return BobData(
        algo=algo, hash_=hash_, max_age=max_age, data=bob_data, cid=cid, type=type_
    )



# Copyright (C) 2020 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from typing import Literal

from dataclasses import dataclass

from nbxmpp.const import Affiliation
from nbxmpp.const import Role
from nbxmpp.modules.vcard_temp import VCard
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import InvalidJid
from nbxmpp.protocol import Iq
from nbxmpp.protocol import JID
from nbxmpp.protocol import Message
from nbxmpp.protocol import StanzaMalformed
from nbxmpp.simplexml import Node
from nbxmpp.structs import DiscoInfo
from nbxmpp.structs import MucUserData

AffiliationT = Literal["owner", "admin", "member", "outcast", "none"]
RoleT = Literal["moderator", "participant", "visitor", "none"]


@dataclass
class MucInfoResult:
    info: DiscoInfo
    vcard: VCard | None = None
    redirected: bool = False


def make_affiliation_request(jid: JID | str, affiliation: AffiliationT) -> Iq:
    iq = Iq(typ="get", to=jid, queryNS=Namespace.MUC_ADMIN)
    item = iq.setQuery().setTag("item")
    item.setAttr("affiliation", affiliation)
    return iq


def make_set_affiliation_request(
    room_jid: JID, users_dict: dict[JID, dict[str, str | None]]
) -> Iq:
    iq = Iq(typ="set", to=room_jid, queryNS=Namespace.MUC_ADMIN)
    item = iq.setQuery()
    for jid, data in users_dict.items():
        affiliation = data.get("affiliation")
        reason = data.get("reason")
        nick = data.get("nick")
        item_tag = item.addChild("item", {"jid": jid, "affiliation": affiliation})
        if reason is not None:
            item_tag.setTagData("reason", reason)

        if nick is not None:
            item_tag.setAttr("nick", nick)

    return iq


def make_destroy_request(room_jid: JID, reason: str | None, jid: str | None) -> Iq:
    iq = Iq(typ="set", queryNS=Namespace.MUC_OWNER, to=room_jid)
    destroy = iq.setQuery().setTag("destroy")

    if reason:
        destroy.setTagData("reason", reason)

    if jid:
        destroy.setAttr("jid", jid)

    return iq


def make_set_config_request(room_jid: JID, form) -> Iq:
    iq = Iq(typ="set", to=room_jid, queryNS=Namespace.MUC_OWNER)
    query = iq.setQuery()
    form.setAttr("type", "submit")
    query.addChild(node=form)
    return iq


def make_config_request(room_jid: JID) -> Iq:
    iq = Iq(typ="get", queryNS=Namespace.MUC_OWNER, to=room_jid)
    return iq


def make_cancel_config_request(room_jid: JID) -> Iq:
    cancel = Node(tag="x", attrs={"xmlns": Namespace.DATA, "type": "cancel"})
    iq = Iq(typ="set", queryNS=Namespace.MUC_OWNER, payload=cancel, to=room_jid)
    return iq


def make_set_role_request(
    room_jid: JID, nick: str, role: RoleT, reason: str | None
) -> Iq:
    iq = Iq(typ="set", to=room_jid, queryNS=Namespace.MUC_ADMIN)
    item = iq.setQuery().setTag("item")
    item.setAttr("nick", nick)
    item.setAttr("role", role)
    if reason:
        item.addChild(name="reason", payload=reason)

    return iq


def make_captcha_request(room_jid: JID, form_node) -> Iq:
    iq = Iq(typ="set", to=room_jid)
    captcha = iq.addChild(name="captcha", namespace=Namespace.CAPTCHA)
    captcha.addChild(node=form_node)
    return iq


def build_direct_invite(
    room: JID, to: JID, reason: str | None, password: str | None, continue_: bool | None
) -> Message:
    message = Message(to=to)
    attrs = {"jid": room}
    if reason:
        attrs["reason"] = reason
    if continue_:
        attrs["continue"] = "true"
    if password:
        attrs["password"] = password
    message.addChild(name="x", attrs=attrs, namespace=Namespace.CONFERENCE)
    return message


def build_mediated_invite(
    room: JID, to: JID, reason: str | None, password: str | None, continue_: bool | None
) -> Message:
    message = Message(to=room)
    muc_user = message.addChild("x", namespace=Namespace.MUC_USER)
    invite = muc_user.addChild("invite", attrs={"to": to})
    if continue_:
        invite.addChild(name="continue")
    if reason:
        invite.setTagData("reason", reason)
    if password:
        muc_user.setTagData("password", password)
    return message


def parse_muc_user(muc_user: Node, is_presence: bool = True) -> MucUserData | None:
    item = muc_user.getTag("item")
    if item is None:
        return None

    item_dict = item.getAttrs()

    role = item_dict.get("role")
    if role is not None:
        try:
            role = Role(role)
        except ValueError:
            raise StanzaMalformed("invalid role %s" % role)

    elif is_presence:
        # role attr MUST be included in all presence broadcasts
        raise StanzaMalformed("role attr missing")

    affiliation = item_dict.get("affiliation")
    if affiliation is not None:
        try:
            affiliation = Affiliation(affiliation)
        except ValueError:
            raise StanzaMalformed("invalid affiliation %s" % affiliation)

    elif is_presence:
        # affiliation attr MUST be included in all presence broadcasts
        raise StanzaMalformed("affiliation attr missing")

    jid = item_dict.get("jid")
    if jid is not None:
        try:
            jid = JID.from_string(jid, force_bare=True)
        except InvalidJid as error:
            raise StanzaMalformed("invalid jid %s, %s" % (jid, error))

    return MucUserData(
        affiliation=affiliation,
        jid=jid,
        nick=item.getAttr("nick"),
        role=role,
        actor=item.getTagAttr("actor", "nick"),
        reason=item.getTagData("reason"),
    )



# Copyright (C) 2018 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from nbxmpp.const import REGISTER_FIELDS
from nbxmpp.errors import MalformedStanzaError
from nbxmpp.errors import StanzaError
from nbxmpp.modules.bits_of_binary import parse_bob_data
from nbxmpp.modules.dataforms import create_field
from nbxmpp.modules.dataforms import extend_form
from nbxmpp.modules.dataforms import FieldT
from nbxmpp.modules.dataforms import SimpleDataForm
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Iq
from nbxmpp.protocol import JID
from nbxmpp.protocol import Protocol
from nbxmpp.structs import RegisterData


def _make_password_change_request(domain: str, username: str, password: str) -> Iq:
    iq = Iq("set", Namespace.REGISTER, to=domain)
    query = iq.getQuery()
    query.setTagData("username", username)
    query.setTagData("password", password)
    return iq


def _make_password_change_with_form(domain: str, form) -> Iq:
    iq = Iq("set", Namespace.REGISTER, to=domain)
    iq.setQueryPayload(form)
    return iq


def _make_register_form(jid: JID, form) -> Iq:
    iq = Iq("set", Namespace.REGISTER, to=jid)
    if form.is_fake_form():
        query = iq.getTag("query")
        for field in form.iter_fields():
            if field.var == "fakeform":
                continue
            query.addChild(field.var, payload=[field.value])
        return iq

    iq.setQueryPayload(form)
    return iq


def _make_unregister_request(jid: JID) -> Iq:
    iq = Iq("set", to=jid)
    query = iq.setQuery()
    query.setNamespace(Namespace.REGISTER)
    query.addChild("remove")
    return iq


def _parse_oob_url(query: Iq) -> str | None:
    oob = query.getTag("x", namespace=Namespace.X_OOB)
    if oob is not None:
        return oob.getTagData("url") or None
    return None


def _parse_form(stanza):
    query = stanza.getTag("query", namespace=Namespace.REGISTER)
    form = query.getTag("x", namespace=Namespace.DATA)
    if form is None:
        return None

    form = extend_form(node=form)
    field = form.vars.get("FORM_TYPE")
    if field is None:
        return None

    # Invalid urn:xmpp:captcha used by ejabberd
    # See https://github.com/processone/ejabberd/issues/3045
    if field.value in ("jabber:iq:register", "urn:xmpp:captcha"):
        return form
    return None


def _parse_fields_form(query: Protocol) -> SimpleDataForm | None:
    fields: list[FieldT] = []
    for field in query.getChildren():
        field_name = field.getName()
        if field_name not in REGISTER_FIELDS:
            continue

        required = field_name in ("username", "password")
        typ = "text-single" if field_name != "password" else "text-private"
        fields.append(create_field(typ=typ, var=field_name, required=required))

    if not fields:
        return None

    fields.append(create_field(typ="hidden", var="fakeform"))
    return SimpleDataForm(
        type_="form", instructions=query.getTagData("instructions"), fields=fields
    )


def _parse_register_data(response: Protocol) -> RegisterData:
    query = response.getTag("query", namespace=Namespace.REGISTER)
    if query is None:
        raise StanzaError(response)

    instructions = query.getTagData("instructions") or None

    data = RegisterData(
        instructions=instructions,
        form=_parse_form(response),
        fields_form=_parse_fields_form(query),
        oob_url=_parse_oob_url(query),
        bob_data=parse_bob_data(query),
    )

    if data.form is None and data.fields_form is None and data.oob_url is None:
        raise MalformedStanzaError("invalid register response", response)

    return data



# Copyright (C) 2019 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from typing import Any
from typing import Literal
from typing import overload
from typing import TYPE_CHECKING

import logging
import time
from collections.abc import Callable
from xml.parsers.expat import ExpatError

from gi.repository import GLib

from nbxmpp.exceptions import StanzaDecrypted
from nbxmpp.modules.activity import Activity
from nbxmpp.modules.adhoc import AdHoc
from nbxmpp.modules.annotations import Annotations
from nbxmpp.modules.attention import Attention
from nbxmpp.modules.blocking import Blocking
from nbxmpp.modules.bookmarks.native_bookmarks import NativeBookmarks
from nbxmpp.modules.bookmarks.pep_bookmarks import PEPBookmarks
from nbxmpp.modules.bookmarks.private_bookmarks import PrivateBookmarks
from nbxmpp.modules.captcha import Captcha
from nbxmpp.modules.chat_markers import ChatMarkers
from nbxmpp.modules.chatstates import Chatstates
from nbxmpp.modules.correction import Correction
from nbxmpp.modules.delay import Delay
from nbxmpp.modules.delimiter import Delimiter
from nbxmpp.modules.discovery import Discovery
from nbxmpp.modules.eme import EME
from nbxmpp.modules.entity_caps import EntityCaps
from nbxmpp.modules.entity_time import EntityTime
from nbxmpp.modules.http_auth import HTTPAuth
from nbxmpp.modules.http_upload import HTTPUpload
from nbxmpp.modules.ibb import IBB
from nbxmpp.modules.idle import Idle
from nbxmpp.modules.iq import BaseIq
from nbxmpp.modules.last_activity import LastActivity
from nbxmpp.modules.location import Location
from nbxmpp.modules.mam import MAM
from nbxmpp.modules.mds import MDS
from nbxmpp.modules.message import BaseMessage
from nbxmpp.modules.misc import unwrap_carbon
from nbxmpp.modules.misc import unwrap_mam
from nbxmpp.modules.mood import Mood
from nbxmpp.modules.muc import MUC
from nbxmpp.modules.muc.hats import Hats
from nbxmpp.modules.muc.moderation import Moderation
from nbxmpp.modules.muclumbus import Muclumbus
from nbxmpp.modules.nickname import Nickname
from nbxmpp.modules.omemo import OMEMO
from nbxmpp.modules.oob import OOB
from nbxmpp.modules.openpgp import OpenPGP
from nbxmpp.modules.pgplegacy import PGPLegacy
from nbxmpp.modules.ping import Ping
from nbxmpp.modules.presence import BasePresence
from nbxmpp.modules.pubsub import PubSub
from nbxmpp.modules.reactions import Reactions
from nbxmpp.modules.receipts import Receipts
from nbxmpp.modules.register import Register
from nbxmpp.modules.replies import Replies
from nbxmpp.modules.retraction import Retraction
from nbxmpp.modules.roster import Roster
from nbxmpp.modules.security_labels import SecurityLabels
from nbxmpp.modules.software_version import SoftwareVersion
from nbxmpp.modules.tune import Tune
from nbxmpp.modules.user_avatar import UserAvatar
from nbxmpp.modules.vcard4 import VCard4
from nbxmpp.modules.vcard_avatar import VCardAvatar
from nbxmpp.modules.vcard_temp import VCardTemp
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import ERR_FEATURE_NOT_IMPLEMENTED
from nbxmpp.protocol import Error
from nbxmpp.protocol import InvalidFrom
from nbxmpp.protocol import InvalidJid
from nbxmpp.protocol import InvalidStanza
from nbxmpp.protocol import Iq
from nbxmpp.protocol import Message
from nbxmpp.protocol import NodeProcessed
from nbxmpp.protocol import Presence
from nbxmpp.protocol import Protocol
from nbxmpp.protocol import StreamErrorNode
from nbxmpp.simplexml import Node
from nbxmpp.simplexml import NodeBuilder
from nbxmpp.structs import StanzaHandler
from nbxmpp.util import get_properties_struct
from nbxmpp.util import INVALID_XML_RX
from nbxmpp.util import is_websocket_close
from nbxmpp.util import is_websocket_stream_error
from nbxmpp.util import LogAdapter
from nbxmpp.util import Observable

if TYPE_CHECKING:
    from nbxmpp.client import Client

log = logging.getLogger("nbxmpp.dispatcher")

NBXMPPModuleNameT = Literal[
    "Activity",
    "Activity",
    "AdHoc",
    "Annotations",
    "Attention",
    "BasePresence",
    "BaseMessage",
    "BaseIq",
    "Blocking",
    "Captcha",
    "ChatMarkers",
    "Chatstates",
    "Correction",
    "Delay",
    "Delimiter",
    "Discovery",
    "EME",
    "EntityCaps",
    "EntityTime",
    "Hats",
    "HTTPAuth",
    "HTTPUpload",
    "IBB",
    "Idle",
    "LastActivity",
    "Location",
    "MAM",
    "MDS",
    "Moderation",
    "Mood",
    "MUC",
    "Muclumbus",
    "NativeBookmarks",
    "Nickname",
    "OMEMO",
    "OOB",
    "OpenPGP",
    "PEPBookmarks",
    "PGPLegacy",
    "Ping",
    "PrivateBookmarks",
    "PubSub",
    "Reactions",
    "Receipts",
    "Register",
    "Replies",
    "Retraction",
    "Roster",
    "SecurityLabels",
    "SoftwareVersion",
    "Tune",
    "UserAvatar",
    "VCardAvatar",
    "VCardTemp",
    "VCard4",
]
NBXMPPModuleT = (
    Activity
    | AdHoc
    | Annotations
    | Attention
    | BasePresence
    | BaseMessage
    | BaseIq
    | Blocking
    | Captcha
    | ChatMarkers
    | Chatstates
    | Correction
    | Delay
    | Delimiter
    | Discovery
    | EME
    | EntityCaps
    | EntityTime
    | Hats
    | HTTPAuth
    | HTTPUpload
    | IBB
    | Idle
    | LastActivity
    | Location
    | MAM
    | MDS
    | Moderation
    | Mood
    | MUC
    | Muclumbus
    | NativeBookmarks
    | Nickname
    | OMEMO
    | OOB
    | OpenPGP
    | PEPBookmarks
    | PGPLegacy
    | Ping
    | PrivateBookmarks
    | PubSub
    | Reactions
    | Receipts
    | Register
    | Replies
    | Retraction
    | Roster
    | SecurityLabels
    | SoftwareVersion
    | Tune
    | UserAvatar
    | VCardAvatar
    | VCardTemp
    | VCard4
)


class StanzaDispatcher(Observable):
    """
    Dispatches stanzas to handlers

    Signals:
        before-dispatch
        parsing-error
        stream-end

    """

    def __init__(self, client: Client) -> None:
        Observable.__init__(self, log)
        self._client = client
        self._modules: dict[str, NBXMPPModuleT] = {}
        self._parser: NodeBuilder | None = None
        self._websocket_stream_error: str | None = None

        self._log = LogAdapter(log, {"context": client.log_context})

        self._handlers: dict[str, dict[str, dict[str, Any]]] = {}

        self._id_callbacks: dict[str, tuple[Callable[..., Any], float | None, Any]] = {}
        self._dispatch_callback: Callable[..., Any] | None = None
        self._timeout_id: int | None = None

        self._stanza_types = {
            "iq": Iq,
            "message": Message,
            "presence": Presence,
            "error": StreamErrorNode,
        }

        self._register_namespace("unknown")
        self._register_namespace(Namespace.STREAMS)
        self._register_namespace(Namespace.CLIENT)
        self._register_protocol("iq", Iq)
        self._register_protocol("presence", Presence)
        self._register_protocol("message", Message)

        self._register_modules()

    def set_dispatch_callback(self, callback: Callable[..., Any]) -> None:
        self._log.info("Set dispatch callback: %s", callback)
        self._dispatch_callback = callback

    @overload
    def get_module(self, name: Literal["Activity"]) -> Activity: ...
    @overload
    def get_module(self, name: Literal["AdHoc"]) -> AdHoc: ...
    @overload
    def get_module(self, name: Literal["Annotations"]) -> Annotations: ...
    @overload
    def get_module(self, name: Literal["Attention"]) -> Attention: ...
    @overload
    def get_module(self, name: Literal["Blocking"]) -> Blocking: ...
    @overload
    def get_module(self, name: Literal["NativeBookmarks"]) -> NativeBookmarks: ...
    @overload
    def get_module(self, name: Literal["PEPBookmarks"]) -> PEPBookmarks: ...
    @overload
    def get_module(self, name: Literal["PrivateBookmarks"]) -> PrivateBookmarks: ...
    @overload
    def get_module(self, name: Literal["Captcha"]) -> Captcha: ...
    @overload
    def get_module(self, name: Literal["ChatMarkers"]) -> ChatMarkers: ...
    @overload
    def get_module(self, name: Literal["Chatstates"]) -> Chatstates: ...
    @overload
    def get_module(self, name: Literal["Correction"]) -> Correction: ...
    @overload
    def get_module(self, name: Literal["Delay"]) -> Delay: ...
    @overload
    def get_module(self, name: Literal["Delimiter"]) -> Delimiter: ...
    @overload
    def get_module(self, name: Literal["Discovery"]) -> Discovery: ...
    @overload
    def get_module(self, name: Literal["EME"]) -> EME: ...
    @overload
    def get_module(self, name: Literal["EntityCaps"]) -> EntityCaps: ...
    @overload
    def get_module(self, name: Literal["EntityTime"]) -> EntityTime: ...
    @overload
    def get_module(self, name: Literal["Hats"]) -> Hats: ...
    @overload
    def get_module(self, name: Literal["HTTPAuth"]) -> HTTPAuth: ...
    @overload
    def get_module(self, name: Literal["HTTPUpload"]) -> HTTPUpload: ...
    @overload
    def get_module(self, name: Literal["IBB"]) -> IBB: ...
    @overload
    def get_module(self, name: Literal["Idle"]) -> Idle: ...
    @overload
    def get_module(self, name: Literal["BaseIq"]) -> BaseIq: ...
    @overload
    def get_module(self, name: Literal["LastActivity"]) -> LastActivity: ...
    @overload
    def get_module(self, name: Literal["Location"]) -> Location: ...
    @overload
    def get_module(self, name: Literal["MAM"]) -> MAM: ...
    @overload
    def get_module(self, name: Literal["MDS"]) -> MDS: ...
    @overload
    def get_module(self, name: Literal["BaseMessage"]) -> BaseMessage: ...
    @overload
    def get_module(self, name: Literal["Mood"]) -> Mood: ...
    @overload
    def get_module(self, name: Literal["MUC"]) -> MUC: ...
    @overload
    def get_module(self, name: Literal["Moderation"]) -> Moderation: ...
    @overload
    def get_module(self, name: Literal["Muclumbus"]) -> Muclumbus: ...
    @overload
    def get_module(self, name: Literal["Nickname"]) -> Nickname: ...
    @overload
    def get_module(self, name: Literal["OMEMO"]) -> OMEMO: ...
    @overload
    def get_module(self, name: Literal["OOB"]) -> OOB: ...
    @overload
    def get_module(self, name: Literal["OpenPGP"]) -> OpenPGP: ...
    @overload
    def get_module(self, name: Literal["PGPLegacy"]) -> PGPLegacy: ...
    @overload
    def get_module(self, name: Literal["Ping"]) -> Ping: ...
    @overload
    def get_module(self, name: Literal["BasePresence"]) -> BasePresence: ...
    @overload
    def get_module(self, name: Literal["PubSub"]) -> PubSub: ...
    @overload
    def get_module(self, name: Literal["Reactions"]) -> Reactions: ...
    @overload
    def get_module(self, name: Literal["Receipts"]) -> Receipts: ...
    @overload
    def get_module(self, name: Literal["Register"]) -> Register: ...
    @overload
    def get_module(self, name: Literal["Replies"]) -> Replies: ...
    @overload
    def get_module(self, name: Literal["Retraction"]) -> Retraction: ...
    @overload
    def get_module(self, name: Literal["Roster"]) -> Roster: ...
    @overload
    def get_module(self, name: Literal["SecurityLabels"]) -> SecurityLabels: ...
    @overload
    def get_module(self, name: Literal["SoftwareVersion"]) -> SoftwareVersion: ...
    @overload
    def get_module(self, name: Literal["Tune"]) -> Tune: ...
    @overload
    def get_module(self, name: Literal["UserAvatar"]) -> UserAvatar: ...
    @overload
    def get_module(self, name: Literal["VCard4"]) -> VCard4: ...
    @overload
    def get_module(self, name: Literal["VCardAvatar"]) -> VCardAvatar: ...
    @overload
    def get_module(self, name: Literal["VCardTemp"]) -> VCardTemp: ...

    def get_module(self, name: NBXMPPModuleNameT) -> NBXMPPModuleT:
        return self._modules[name]

    def _register_modules(self):
        assert self._client is not None
        self._modules["Activity"] = Activity(self._client)
        self._modules["AdHoc"] = AdHoc(self._client)
        self._modules["Annotations"] = Annotations(self._client)
        self._modules["Attention"] = Attention(self._client)
        self._modules["BasePresence"] = BasePresence(self._client)
        self._modules["BaseMessage"] = BaseMessage(self._client)
        self._modules["BaseIq"] = BaseIq(self._client)
        self._modules["Blocking"] = Blocking(self._client)
        self._modules["Captcha"] = Captcha(self._client)
        self._modules["ChatMarkers"] = ChatMarkers(self._client)
        self._modules["Chatstates"] = Chatstates(self._client)
        self._modules["Correction"] = Correction(self._client)
        self._modules["Delay"] = Delay(self._client)
        self._modules["Delimiter"] = Delimiter(self._client)
        self._modules["Discovery"] = Discovery(self._client)
        self._modules["EME"] = EME(self._client)
        self._modules["EntityCaps"] = EntityCaps(self._client)
        self._modules["EntityTime"] = EntityTime(self._client)
        self._modules["Hats"] = Hats(self._client)
        self._modules["HTTPAuth"] = HTTPAuth(self._client)
        self._modules["HTTPUpload"] = HTTPUpload(self._client)
        self._modules["IBB"] = IBB(self._client)
        self._modules["Idle"] = Idle(self._client)
        self._modules["LastActivity"] = LastActivity(self._client)
        self._modules["Location"] = Location(self._client)
        self._modules["MAM"] = MAM(self._client)
        self._modules["MDS"] = MDS(self._client)
        self._modules["Moderation"] = Moderation(self._client)
        self._modules["Mood"] = Mood(self._client)
        self._modules["MUC"] = MUC(self._client)
        self._modules["Muclumbus"] = Muclumbus(self._client)
        self._modules["NativeBookmarks"] = NativeBookmarks(self._client)
        self._modules["Nickname"] = Nickname(self._client)
        self._modules["OMEMO"] = OMEMO(self._client)
        self._modules["OOB"] = OOB(self._client)
        self._modules["OpenPGP"] = OpenPGP(self._client)
        self._modules["PEPBookmarks"] = PEPBookmarks(self._client)
        self._modules["PGPLegacy"] = PGPLegacy(self._client)
        self._modules["Ping"] = Ping(self._client)
        self._modules["PrivateBookmarks"] = PrivateBookmarks(self._client)
        self._modules["PubSub"] = PubSub(self._client)
        self._modules["Reactions"] = Reactions(self._client)
        self._modules["Receipts"] = Receipts(self._client)
        self._modules["Register"] = Register(self._client)
        self._modules["Replies"] = Replies(self._client)
        self._modules["Retraction"] = Retraction(self._client)
        self._modules["Roster"] = Roster(self._client)
        self._modules["SecurityLabels"] = SecurityLabels(self._client)
        self._modules["SoftwareVersion"] = SoftwareVersion(self._client)
        self._modules["Tune"] = Tune(self._client)
        self._modules["UserAvatar"] = UserAvatar(self._client)
        self._modules["VCardAvatar"] = VCardAvatar(self._client)
        self._modules["VCardTemp"] = VCardTemp(self._client)
        self._modules["VCard4"] = VCard4(self._client)

        for instance in self._modules.values():
            for handler in instance.handlers:
                self.register_handler(handler)

    def reset_parser(self) -> None:
        if self._parser is not None:
            self._parser.dispatch = None
            self._parser.destroy()
            self._parser = None

        self._parser = NodeBuilder(dispatch_depth=2, finished=False)
        self._parser.dispatch = self.dispatch

    def replace_non_character(self, data: str) -> str:
        return INVALID_XML_RX.sub("\ufffd", data)

    def process_data(self, data: str) -> None:
        # Parse incoming data

        data = self.replace_non_character(data)

        if self._client.is_websocket:
            stanza = Node(node=data)
            if is_websocket_stream_error(stanza):
                for tag in stanza.getChildren():
                    name = tag.getName()
                    if name != "text" and tag.getNamespace() == Namespace.XMPP_STREAMS:
                        self._websocket_stream_error = name

            elif is_websocket_close(stanza):
                self._log.info("Stream <close> received")
                self.notify("stream-end", self._websocket_stream_error)
                return

            self.dispatch(stanza)
            return

        try:
            self._parser.Parse(data)
        except (ExpatError, ValueError) as error:
            self._log.error("XML parsing error: %s", error)
            self.notify("parsing-error", str(error))
            return

        # end stream:stream tag received
        if self._parser.has_received_endtag():
            self._log.info("End of stream: %s", self._parser.stream_error)
            self.notify("stream-end", self._parser.stream_error)
            return

    def _register_namespace(self, xmlns: str) -> None:
        """
        Setup handler structure for namespace
        """
        self._log.debug('Register namespace "%s"', xmlns)
        self._handlers[xmlns] = {}
        self._register_protocol("error", Protocol, xmlns=xmlns)
        self._register_protocol("unknown", Protocol, xmlns=xmlns)
        self._register_protocol("default", Protocol, xmlns=xmlns)

    def _register_protocol(
        self, tag_name: str, protocol: Any, xmlns: str | None = None
    ) -> None:
        """
        Register protocol for top level tag names
        """
        if xmlns is None:
            xmlns = Namespace.CLIENT
        self._log.debug('Register protocol "%s (%s)" as %s', tag_name, xmlns, protocol)
        self._handlers[xmlns][tag_name] = {"type": protocol, "default": []}

    def register_handler(self, handler: StanzaHandler) -> None:
        """
        Register handler
        """

        xmlns = handler.xmlns or Namespace.CLIENT

        typ = handler.typ
        if not typ and not handler.ns:
            typ = "default"

        self._log.debug(
            'Register handler %s for "%s" type->%s ns->%s(%s) priority->%s',
            handler.callback,
            handler.name,
            typ,
            handler.ns,
            xmlns,
            handler.priority,
        )

        if xmlns not in self._handlers:
            self._register_namespace(xmlns)
        if handler.name not in self._handlers[xmlns]:
            self._register_protocol(handler.name, Protocol, xmlns)

        specific = typ + handler.ns
        if specific not in self._handlers[xmlns][handler.name]:
            self._handlers[xmlns][handler.name][specific] = []

        self._handlers[xmlns][handler.name][specific].append(
            {
                "func": handler.callback,
                "priority": handler.priority,
                "specific": specific,
            }
        )

    def unregister_handler(self, handler: StanzaHandler) -> None:
        """
        Unregister handler
        """

        xmlns = handler.xmlns or Namespace.CLIENT

        typ = handler.typ
        if not typ and not handler.ns:
            typ = "default"

        specific = typ + handler.ns
        try:
            self._handlers[xmlns][handler.name][specific]
        except KeyError:
            return

        for handler_dict in self._handlers[xmlns][handler.name][specific]:
            if handler_dict["func"] != handler.callback:
                continue

            try:
                self._handlers[xmlns][handler.name][specific].remove(handler_dict)
            except ValueError:
                self._log.warning(
                    'Unregister failed: %s for "%s" type->%s ns->%s(%s)',
                    handler.callback,
                    handler.name,
                    typ,
                    handler.ns,
                    xmlns,
                )
            else:
                self._log.debug(
                    'Unregister handler %s for "%s" type->%s ns->%s(%s)',
                    handler.callback,
                    handler.name,
                    typ,
                    handler.ns,
                    xmlns,
                )

    def _default_handler(self, stanza: Protocol) -> None:
        """
        Return stanza back to the sender with <feature-not-implemented/> error
        """
        if stanza.getType() in ("get", "set"):
            self._client.send_stanza(Error(stanza, ERR_FEATURE_NOT_IMPLEMENTED))

    def dispatch(self, stanza: Protocol) -> None:
        self.notify("before-dispatch", stanza)

        if self._dispatch_callback is not None:
            name = stanza.getName()
            protocol_class = self._stanza_types.get(name)
            if protocol_class is not None:
                stanza = protocol_class(node=stanza)
            self._dispatch_callback(stanza)
            return

        # Count stanza
        self._client._smacks.count_incoming(stanza.getName())

        name = stanza.getName()
        xmlns = stanza.getNamespace()

        if xmlns not in self._handlers:
            self._log.warning("Unknown namespace: %s", xmlns)
            xmlns = "unknown"

        if name not in self._handlers[xmlns]:
            self._log.warning("Unknown stanza: %s", stanza)
            name = "unknown"

        # Convert simplexml to Protocol object
        try:
            stanza = self._handlers[xmlns][name]["type"](node=stanza)
        except InvalidJid:
            self._log.warning("Invalid JID, ignoring stanza")
            self._log.warning(stanza)
            return

        own_jid = self._client.get_bound_jid()
        properties = get_properties_struct(name, own_jid)

        if name == "iq":
            if stanza.getFrom() is None and own_jid is not None:
                stanza.setFrom(own_jid.bare)

        if name == "message":
            # https://tools.ietf.org/html/rfc6120#section-8.1.1.1
            # If the stanza does not include a 'to' address then the client MUST
            # treat it as if the 'to' address were included with a value of the
            # client's full JID.

            to = stanza.getTo()
            if to is None:
                stanza.setTo(own_jid)

            elif not to.bare_match(own_jid):
                self._log.warning("Message addressed to someone else: %s", stanza)
                return

            if stanza.getFrom() is None:
                stanza.setFrom(own_jid.bare)

            # Unwrap carbon
            try:
                stanza, properties.carbon = unwrap_carbon(stanza, own_jid)
            except (InvalidFrom, InvalidJid) as exc:
                self._log.warning(exc)
                self._log.warning(stanza)
                return
            except NodeProcessed as exc:
                self._log.info(exc)
                return

            # Unwrap mam
            try:
                stanza, properties.mam = unwrap_mam(stanza, own_jid)
            except (InvalidStanza, InvalidJid) as exc:
                self._log.warning(exc)
                self._log.warning(stanza)
                return

        typ = stanza.getType()
        if not typ:
            if name == "message":
                typ = "normal"
            elif name == "presence":
                typ = "available"
            else:
                typ = ""

        # Process callbacks
        _id = stanza.getID()
        func, _timeout, user_data = self._id_callbacks.pop(_id, (None, None, {}))
        if user_data is None:
            user_data = {}

        if func is not None:
            try:
                func(self._client, stanza, **user_data)
            except Exception:
                self._log.exception("Error while handling stanza")
            return

        props = stanza.getProperties()
        self._log.debug("type: %s, properties: %s", typ, props)

        chain = self._build_handler_chain(xmlns, name, typ, props)

        try:
            self._execute_handler_chain(chain, stanza, properties)
        except StanzaDecrypted:
            props = stanza.getProperties()
            self._log.debug("type: %s, properties after decryption: %s", typ, props)
            chain = self._build_handler_chain(
                xmlns, name, typ, props, after_decryption=True
            )
            self._execute_handler_chain(chain, stanza, properties)

    def _build_handler_chain(
        self,
        xmlns: str,
        name: str,
        typ: str,
        props: Any,
        *,
        after_decryption: bool = False,
    ) -> list[dict[str, Any]]:

        # Gather specifics depending on stanza properties
        specifics = ["default"]
        if typ and typ in self._handlers[xmlns][name]:
            specifics.append(typ)

        for prop in props:
            if prop in self._handlers[xmlns][name]:
                specifics.append(prop)

            if typ and typ + prop in self._handlers[xmlns][name]:
                specifics.append(typ + prop)

        # Create the handler chain
        chain: list[dict[str, Any]] = []
        chain += self._handlers[xmlns]["default"]["default"]
        for specific in specifics:
            chain += self._handlers[xmlns][name][specific]

        # Sort chain with priority
        chain.sort(key=lambda x: x["priority"])

        if after_decryption:
            # Filter everything out which was executed before decryption
            # so it is not executed again
            chain = list(filter(lambda x: x["priority"] > 9, chain))

        return chain

    def _execute_handler_chain(
        self, chain: Any, stanza: Protocol, properties: Any
    ) -> None:

        for handler in chain:
            self._log.info("Call handler: %s", handler["func"].__qualname__)
            try:
                handler["func"](self._client, stanza, properties)
            except NodeProcessed:
                return
            except StanzaDecrypted:
                raise
            except Exception:
                self._log.exception("Handler exception:")
                return

        # Stanza was not processed call default handler
        self._default_handler(stanza)

    def add_callback_for_id(
        self, id_: str, func: Callable[..., Any], timeout: float | None, user_data: Any
    ) -> None:
        if timeout is not None and self._timeout_id is None:
            self._log.info("Add timeout check")
            self._timeout_id = GLib.timeout_add_seconds(1, self._timeout_check)
            timeout = time.monotonic() + timeout
        self._id_callbacks[id_] = (func, timeout, user_data)

    def _timeout_check(self) -> bool:
        self._log.info("Run timeout check")
        timeouts: dict[str, tuple[Callable[..., Any], float | None, Any]] = {}
        for id_, data in self._id_callbacks.items():
            if data[1] is not None:
                timeouts[id_] = data

        if not timeouts:
            self._log.info("Remove timeout check, no timeouts scheduled")
            self._timeout_id = None
            return False

        for id_, data in timeouts.items():
            func, timeout, user_data = data

            if user_data is None:
                user_data = {}

            if timeout < time.monotonic():
                self._id_callbacks.pop(id_)
                func(self._client, None, **user_data)
        return True

    def _remove_timeout_source(self) -> None:
        if self._timeout_id is not None:
            GLib.source_remove(self._timeout_id)
            self._timeout_id = None

    def remove_iq_callback(self, id_: str) -> None:
        self._id_callbacks.pop(id_, None)

    def clear_iq_callbacks(self) -> None:
        self._log.info("Clear IQ callbacks")
        self._id_callbacks.clear()

    def cleanup(self) -> None:
        self._client = None
        self._modules = {}
        self._parser = None
        self.clear_iq_callbacks()
        self._dispatch_callback = None
        self._handlers.clear()
        self._remove_timeout_source()
        self.remove_subscriptions()



# Copyright (C) 2018 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from typing import TYPE_CHECKING

from nbxmpp.modules.base import BaseModule
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Message
from nbxmpp.structs import ChatMarker
from nbxmpp.structs import MessageProperties
from nbxmpp.structs import StanzaHandler

if TYPE_CHECKING:
    from nbxmpp.client import Client


class ChatMarkers(BaseModule):
    def __init__(self, client: Client) -> None:
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = [
            StanzaHandler(
                name="message",
                callback=self._process_message_marker,
                ns=Namespace.CHATMARKERS,
                priority=15,
            ),
        ]

    def _process_message_marker(
        self, _client: Client, stanza: Message, properties: MessageProperties
    ) -> None:
        type_ = stanza.getTag("received", namespace=Namespace.CHATMARKERS)
        if type_ is None:
            type_ = stanza.getTag("displayed", namespace=Namespace.CHATMARKERS)
            if type_ is None:
                type_ = stanza.getTag("acknowledged", namespace=Namespace.CHATMARKERS)
                if type_ is None:
                    return

        name = type_.getName()
        id_ = type_.getAttr("id")
        if id_ is None:
            self._log.warning("Chatmarker without id")
            self._log.warning(stanza)
            return

        properties.marker = ChatMarker(name, id_)



# Copyright (C) 2020 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from typing import Any
from typing import Literal
from typing import TYPE_CHECKING

import logging
from collections import defaultdict
from collections.abc import ValuesView
from dataclasses import dataclass
from dataclasses import field

from nbxmpp.errors import MalformedStanzaError
from nbxmpp.modules.base import BaseModule
from nbxmpp.modules.util import finalize
from nbxmpp.modules.util import raise_if_error
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import JID
from nbxmpp.simplexml import Node
from nbxmpp.task import iq_request_task

if TYPE_CHECKING:
    from nbxmpp.client import Client

log = logging.getLogger("nbxmpp.m.vcard4")


ALLOWED_SEX_VALUES = ["M", "F", "O", "N", "U"]
ALLOWED_KIND_VALUES = ["individual", "group", "org", "location"]

# Cardinality
# 1     Exactly one instance per vCard MUST be present.
# *1    Exactly one instance per vCard MAY be present.
# 1*    One or more instances per vCard MUST be present.
# *     One or more instances per vCard MAY be present.

PROPERTY_DEFINITION: dict[str, tuple[list[str], Literal["*", "*1", "1*", "1"]]] = {
    "source": (["altid", "pid", "pref", "mediatype"], "*"),
    "kind": ([], "*1"),
    "fn": (["language", "altid", "pid", "pref", "type"], "1*"),
    "n": (["language", "altid", "sort-as"], "*1"),
    "nickname": (["language", "altid", "pid", "pref", "type"], "*"),
    "photo": (["altid", "pid", "pref", "type", "mediatype"], "*"),
    "bday": (["altid", "calscale"], "*1"),
    "anniversary": (["altid", "calscale"], "*1"),
    "gender": ([], "*1"),
    "adr": (["language", "altid", "pid", "pref", "type", "geo", "tz", "label"], "*"),
    "tel": (["altid", "pid", "pref", "type", "mediatype"], "*"),
    "email": (["altid", "pid", "pref", "type"], "*"),
    "impp": (["altid", "pid", "pref", "type", "mediatype"], "*"),
    "lang": (["altid", "pid", "pref", "type"], "*"),
    "tz": (["altid", "pid", "pref", "type", "mediatype"], "*"),
    "geo": (["altid", "pid", "pref", "type", "mediatype"], "*"),
    "title": (["language", "altid", "pid", "pref", "type"], "*"),
    "role": (["language", "altid", "pid", "pref", "type"], "*"),
    "logo": (["language", "altid", "pid", "pref", "type", "mediatype"], "*"),
    "org": (["language", "altid", "pid", "pref", "type", "sort-as"], "*"),
    "member": (["altid", "pid", "pref", "mediatype"], "*"),
    "related": (["altid", "pid", "pref", "type", "mediatype"], "*"),
    "categories": (["altid", "pid", "pref", "type"], "*"),
    "note": (["language", "altid", "pid", "pref", "type"], "*"),
    "prodid": ([], "*1"),
    "rev": ([], "*1"),
    "sound": (["language", "altid", "pid", "pref", "type", "mediatype"], "*"),
    "uid": ([], "*1"),
    "clientpidmap": ([], "*"),
    "url": (["altid", "pid", "pref", "type", "mediatype"], "*"),
    "key": (["altid", "pid", "pref", "type", "mediatype"], "*"),
    "fburl": (["altid", "pid", "pref", "type", "mediatype"], "*"),
    "caladruri": (["altid", "pid", "pref", "type", "mediatype"], "*"),
    "caluri": (["altid", "pid", "pref", "type", "mediatype"], "*"),
}


PROPERTY_VALUE_TYPES = {
    "bday": ["date", "time", "date-time", "text"],
    "anniversary": ["date", "time", "date-time", "text"],
    "key": ["text", "uri"],
    "tel": ["uri", "text"],
    "tz": ["text", "uri", "utc-offset"],
    "related": ["text", "uri"],
}


def get_data_from_children(node: Node, child_name: str) -> list[str]:
    values: list[str] = []
    child_nodes = node.getTags(child_name)
    for child_node in child_nodes:
        child_value = child_node.getData()
        if child_value:
            values.append(child_value)
    return values


def add_children(node: Node, child_name: str, values: list[str]) -> None:
    for value in values:
        node.addChild(child_name, payload=value)


def get_multiple_type_value(node: Node, types: list[str]) -> tuple[str, str]:
    for type_ in types:
        value = node.getTagData(type_)
        if value:
            return type_, value

    raise ValueError("no value found")


def make_parameters(parameters: Parameters) -> Node:
    parameters_node = Node("parameters")
    for parameter in parameters.values():
        parameters_node.addChild(node=parameter.to_node())
    return parameters_node


def get_parameters(node: Node) -> Parameters:
    name = node.getName()
    definition = PROPERTY_DEFINITION[name]
    allowed_parameters = definition[0]
    parameters_node = node.getTag("parameters")
    if parameters_node is None:
        return Parameters()

    parameters: dict[str, Any] = {}
    for parameter in allowed_parameters:
        parameter_node = parameters_node.getTag(parameter)
        if parameter_node is None:
            continue

        parameter_class = PARAMETER_CLASSES.get(parameter_node.getName())
        if parameter_class is None:
            continue

        parameter = parameter_class.from_node(parameter_node)
        parameters[parameter.name] = parameter

    return Parameters(parameters)


@dataclass
class Parameter:

    name: str
    type: str
    value: str

    @classmethod
    def from_node(cls, node: Node) -> Parameter:
        name = node.getName()
        if name != cls.name:
            raise ValueError(f"invalid parameter name: {name}")

        value = node.getTagData(cls.type)
        if not value:
            raise ValueError("no parameter value found")

        return cls(value)  # type: ignore

    def to_node(self) -> Node:
        node = Node(self.name)
        node.addChild(self.type, payload=self.value)
        return node

    def copy(self) -> Parameter:
        return self.__class__(value=self.value)  # type: ignore


@dataclass
class MultiParameter:

    name: str
    type: str
    values: set[str]

    @classmethod
    def from_node(cls, node: Node) -> MultiParameter:
        name = node.getName()
        if name != cls.name:
            raise ValueError(f"invalid parameter name: {name}")

        value_nodes = node.getTags(cls.type)
        if not value_nodes:
            raise ValueError("no parameter value found")

        values: set[str] = set()
        for value_node in value_nodes:
            value = value_node.getData()
            if value:
                values.add(value)

        if not values:
            raise ValueError("no parameter value found")

        return cls(values)  # type: ignore

    def to_node(self) -> Node:
        node = Node(self.name)
        for value in self.values:
            node.addChild(self.type, payload=value)
        return node

    def copy(self) -> MultiParameter:
        return self.__class__(values=set(self.values))  # type: ignore


@dataclass
class LanguageParameter(Parameter):

    name: str = field(default="language", init=False)
    type: str = field(default="language-tag", init=False)


@dataclass
class PrefParameter(Parameter):

    name: str = field(default="pref", init=False)
    type: str = field(default="integer", init=False)


@dataclass
class AltidParameter(Parameter):

    name: str = field(default="altid", init=False)
    type: str = field(default="text", init=False)


@dataclass
class PidParameter(MultiParameter):

    name: str = field(default="pid", init=False)
    type: str = field(default="text", init=False)


@dataclass
class TypeParameter(MultiParameter):

    name: str = field(default="type", init=False)
    type: str = field(default="text", init=False)


@dataclass
class MediatypeParameter(Parameter):

    name: str = field(default="mediatype", init=False)
    type: str = field(default="text", init=False)


@dataclass
class CalscaleParameter(Parameter):

    name: str = field(default="calscale", init=False)
    type: str = field(default="text", init=False)


@dataclass
class SortasParameter(MultiParameter):

    name: str = field(default="sort-as", init=False)
    type: str = field(default="text", init=False)


@dataclass
class GeoParameter(Parameter):

    name: str = field(default="geo", init=False)
    type: str = field(default="uri", init=False)


@dataclass
class TzParameter:

    name: str = field(default="tz", init=False)
    value_type: str
    value: str

    @classmethod
    def from_node(cls, node: Node) -> TzParameter:
        name = node.getName()
        if name != cls.name:
            raise ValueError(f"invalid property name: {name}")

        value_type, value = get_multiple_type_value(node, ["text", "uri"])
        return cls(value_type, value)

    def to_node(self) -> Node:
        node = Node(self.name)
        node.addChild(self.value_type, payload=self.value)
        return node

    def copy(self) -> TzParameter:
        return self.__class__(value_type=self.value_type, value=self.value)


class Parameters:
    def __init__(
        self,
        parameters: dict[str, Parameter | MultiParameter | TzParameter] | None = None,
    ) -> None:
        if parameters is None:
            parameters = {}
        self._parameters = parameters

    def values(self) -> ValuesView[Parameter | MultiParameter | TzParameter]:
        return self._parameters.values()

    def get_types(self) -> set[str]:
        parameter = self._parameters.get("type")
        if parameter is None:
            return set()

        assert isinstance(parameter, TypeParameter)
        return parameter.values

    def remove_types(self, types: list[str]) -> None:
        parameter = self._parameters.get("type")
        if parameter is None:
            raise ValueError("no type parameter")

        assert isinstance(parameter, TypeParameter)
        for type_ in types:
            parameter.values.discard(type_)

        if not parameter.values:
            self._parameters.pop("type")

    def add_types(self, types: list[str]) -> None:
        parameter = self._parameters.get("type")
        if parameter is None:
            parameter = TypeParameter(set(types))
            self._parameters["type"] = parameter
            return

        assert isinstance(parameter, TypeParameter)
        parameter.values.update(types)

    def copy(self) -> Parameters:
        parameters: dict[str, Parameter | MultiParameter | TzParameter] = {}
        for name, parameter in self._parameters.items():
            parameters[name] = parameter.copy()
        return self.__class__(parameters=parameters)


PARAMETER_CLASSES = {
    "language": LanguageParameter,
    "pref": PrefParameter,
    "altid": AltidParameter,
    "pid": PidParameter,
    "type": TypeParameter,
    "mediatype": MediatypeParameter,
    "calscale": CalscaleParameter,
    "sort-as": SortasParameter,
    "geo": GeoParameter,
    "tz": TzParameter,
}


@dataclass
class UriProperty:

    name: str
    value: str
    parameters: Parameters = field(default_factory=Parameters)

    @classmethod
    def from_node(cls, node: Node) -> UriProperty:
        name = node.getName()
        if name != cls.name:
            raise ValueError(f"invalid property name: {name}")

        value = node.getTagData("uri")
        if not value:
            raise ValueError("no value found")

        parameters = get_parameters(node)

        return cls(value, parameters)  # type: ignore

    def to_node(self) -> Node:
        node = Node(self.name)
        if self.parameters:
            node.addChild(node=make_parameters(self.parameters))
        node.addChild("uri", payload=self.value)  # type: ignore
        return node

    @property
    def is_empty(self) -> bool:
        return not self.value

    def copy(self) -> UriProperty:
        return self.__class__(
            value=self.value, parameters=self.parameters.copy()
        )  # type: ignore


@dataclass
class TextProperty:

    name: str
    value: str
    parameters: Parameters = field(default_factory=Parameters)

    @classmethod
    def from_node(cls, node: Node) -> TelProperty:
        name = node.getName()
        if name != cls.name:
            raise ValueError(f"invalid property name: {name}")

        text = node.getTagData("text")
        if not text:
            raise ValueError("no value found")

        parameters = get_parameters(node)

        return cls(text, parameters)  # type: ignore

    def to_node(self) -> Node:
        node = Node(self.name)
        if self.parameters:
            node.addChild(node=make_parameters(self.parameters))
        node.addChild("text", payload=self.value)
        return node

    @property
    def is_empty(self) -> bool:
        return not self.value

    def copy(self) -> TextProperty:
        return self.__class__(
            value=self.value, parameters=self.parameters.copy()
        )  # type: ignore


@dataclass
class TextListProperty:

    name: str
    values: list[str]
    parameters: Parameters = field(default_factory=Parameters)

    @classmethod
    def from_node(cls, node: Node) -> TextListProperty:
        name = node.getName()
        if name != cls.name:
            raise ValueError(f"invalid property name: {name}")

        text_nodes = node.getTags("text")
        if not text_nodes:
            raise ValueError("no value found")

        values = get_data_from_children(node, "text")
        if not values:
            raise ValueError("no values found")

        parameters = get_parameters(node)

        return cls(values, parameters)  # type: ignore

    def to_node(self) -> Node:
        node = Node(self.name)
        if self.parameters:
            node.addChild(node=make_parameters(self.parameters))
        add_children(node, "text", self.values)
        return node

    @property
    def is_empty(self) -> bool:
        return not self.values

    def copy(self) -> TextListProperty:
        return self.__class__(
            values=list(self.values), parameters=self.parameters.copy()
        )  # type: ignore


@dataclass
class MultipleValueProperty:

    name: str
    value_type: str
    value: str
    parameters: Parameters = field(default_factory=Parameters)

    @classmethod
    def from_node(cls, node: Node) -> MultipleValueProperty:
        name = node.getName()
        if name != cls.name:
            raise ValueError(f"invalid property name: {name}")

        types = PROPERTY_VALUE_TYPES[cls.name]
        value_type, value = get_multiple_type_value(node, types)

        parameters = get_parameters(node)

        return cls(value_type, value, parameters)  # type: ignore

    def to_node(self) -> Node:
        node = Node(self.name)
        if self.parameters:
            node.addChild(node=make_parameters(self.parameters))
        node.addChild(self.value_type, payload=self.value)
        return node

    @property
    def is_empty(self) -> bool:
        return not self.value

    def copy(self) -> MultipleValueProperty:
        return self.__class__(
            value_type=self.value_type,
            value=self.value,
            parameters=self.parameters.copy(),
        )  # type: ignore


@dataclass
class SourceProperty(UriProperty):

    name: str = field(default="source", init=False)


@dataclass
class KindProperty(TextProperty):

    name: str = field(default="kind", init=False)

    @classmethod
    def from_node(cls, node: Node) -> KindProperty:  # type: ignore
        name = node.getName()
        if name != cls.name:
            raise ValueError(f"invalid property name: {name}")

        text = node.getTagData("text")
        if not text:
            raise ValueError("no value found")

        if text not in ALLOWED_KIND_VALUES:
            text = "individual"

        parameters = get_parameters(node)

        return cls(text, parameters)


@dataclass
class FnProperty(TextProperty):

    name: str = field(default="fn", init=False)


@dataclass
class NProperty:

    name: str = field(default="n", init=False)
    surname: list[str] = field(default_factory=list)
    given: list[str] = field(default_factory=list)
    additional: list[str] = field(default_factory=list)
    prefix: list[str] = field(default_factory=list)
    suffix: list[str] = field(default_factory=list)
    parameters: Parameters = field(default_factory=Parameters)

    @classmethod
    def from_node(cls, node: Node) -> NProperty:
        name = node.getName()
        if name != cls.name:
            raise ValueError(f"invalid property name: {name}")

        surname = get_data_from_children(node, "surname")
        given = get_data_from_children(node, "given")
        additional = get_data_from_children(node, "additional")
        prefix = get_data_from_children(node, "prefix")
        suffix = get_data_from_children(node, "suffix")

        parameters = get_parameters(node)

        return cls(surname, given, additional, prefix, suffix, parameters)

    def to_node(self) -> Node:
        node = Node(self.name)
        if self.parameters:
            node.addChild(node=make_parameters(self.parameters))
        add_children(node, "surname", self.surname)
        add_children(node, "given", self.given)
        add_children(node, "additional", self.additional)
        add_children(node, "prefix", self.prefix)
        add_children(node, "suffix", self.suffix)
        return node

    @property
    def is_empty(self) -> bool:
        return not (
            self.surname or self.given or self.additional or self.suffix or self.prefix
        )

    def copy(self) -> NProperty:
        return self.__class__(
            surname=list(self.surname),
            given=list(self.given),
            additional=list(self.additional),
            prefix=list(self.prefix),
            suffix=list(self.suffix),
            parameters=self.parameters.copy(),
        )


@dataclass
class NicknameProperty(TextListProperty):

    name: str = field(default="nickname", init=False)


@dataclass
class PhotoProperty(UriProperty):

    name: str = field(default="photo", init=False)


@dataclass
class BDayProperty(MultipleValueProperty):

    name: str = field(default="bday", init=False)


@dataclass
class AnniversaryProperty(MultipleValueProperty):

    name: str = field(default="anniversary", init=False)


@dataclass
class GenderProperty:

    name: str = field(default="gender", init=False)
    sex: str | None = None
    identity: str | None = None
    parameters: Parameters = field(default_factory=Parameters)

    @classmethod
    def from_node(cls, node: Node) -> GenderProperty:
        name = node.getName()
        if name != cls.name:
            raise ValueError(f"invalid property name: {name}")

        sex = node.getTagData("sex")
        if sex not in ALLOWED_SEX_VALUES:
            sex = None

        identity = node.getTagData("identity")
        if not identity:
            identity = None

        parameters = get_parameters(node)

        return cls(sex, identity, parameters)

    def to_node(self) -> Node:
        node = Node(self.name)
        if self.parameters:
            node.addChild(node=make_parameters(self.parameters))
        if self.sex:
            node.addChild("sex", payload=self.sex)
        if self.identity:
            node.addChild("identity", payload=self.identity)
        return node

    @property
    def is_empty(self) -> bool:
        return not (self.sex or self.identity)

    def copy(self) -> GenderProperty:
        return self.__class__(
            sex=self.sex, identity=self.identity, parameters=self.parameters.copy()
        )


@dataclass
class AdrProperty:

    name: str = field(default="adr", init=False)
    pobox: list[str] = field(default_factory=list)
    ext: list[str] = field(default_factory=list)
    street: list[str] = field(default_factory=list)
    locality: list[str] = field(default_factory=list)
    region: list[str] = field(default_factory=list)
    code: list[str] = field(default_factory=list)
    country: list[str] = field(default_factory=list)
    parameters: Parameters = field(default_factory=Parameters)

    @classmethod
    def from_node(cls, node: Node) -> AdrProperty:
        name = node.getName()
        if name != cls.name:
            raise ValueError(f"invalid property name: {name}")

        pobox = get_data_from_children(node, "pobox")
        ext = get_data_from_children(node, "ext")
        street = get_data_from_children(node, "street")
        locality = get_data_from_children(node, "locality")
        region = get_data_from_children(node, "region")
        code = get_data_from_children(node, "code")
        country = get_data_from_children(node, "country")

        parameters = get_parameters(node)

        return cls(pobox, ext, street, locality, region, code, country, parameters)

    def to_node(self) -> Node:
        node = Node(self.name)
        if self.parameters:
            node.addChild(node=make_parameters(self.parameters))
        add_children(node, "pobox", self.pobox)
        add_children(node, "ext", self.ext)
        add_children(node, "street", self.street)
        add_children(node, "locality", self.locality)
        add_children(node, "region", self.region)
        add_children(node, "code", self.code)
        add_children(node, "country", self.country)
        return node

    @property
    def is_empty(self) -> bool:
        return not (
            self.pobox
            or self.ext
            or self.street
            or self.locality
            or self.region
            or self.code
            or self.country
        )

    def copy(self) -> AdrProperty:
        return self.__class__(
            pobox=list(self.pobox),
            ext=list(self.ext),
            street=list(self.street),
            locality=list(self.locality),
            region=list(self.region),
            code=list(self.code),
            country=list(self.country),
            parameters=self.parameters.copy(),
        )


@dataclass
class TelProperty(MultipleValueProperty):

    name: str = field(default="tel", init=False)


@dataclass
class EmailProperty(TextProperty):

    name: str = field(default="email", init=False)


@dataclass
class ImppProperty(UriProperty):

    name: str = field(default="impp", init=False)


@dataclass
class LangProperty:

    name: str = field(default="lang", init=False)
    value: str
    parameters: Parameters = field(default_factory=Parameters)

    @classmethod
    def from_node(cls, node: Node) -> LangProperty:
        name = node.getName()
        if name != cls.name:
            raise ValueError(f"invalid property name: {name}")

        value = node.getTagData("language-tag")
        if not value:
            raise ValueError("no value found")

        parameters = get_parameters(node)

        return cls(value, parameters)

    def to_node(self) -> Node:
        node = Node(self.name)
        if self.parameters:
            node.addChild(node=make_parameters(self.parameters))
        node.addChild("language-tag", payload=self.value)
        return node

    @property
    def is_empty(self) -> bool:
        return not self.value

    def copy(self) -> LangProperty:
        return self.__class__(value=self.value, parameters=self.parameters.copy())


@dataclass
class TzProperty(MultipleValueProperty):

    name: str = field(default="tz", init=False)


@dataclass
class GeoProperty(UriProperty):

    name: str = field(default="geo", init=False)


@dataclass
class TitleProperty(TextProperty):

    name: str = field(default="title", init=False)


@dataclass
class RoleProperty(TextProperty):

    name: str = field(default="role", init=False)


@dataclass
class LogoProperty(UriProperty):

    name: str = field(default="logo", init=False)


@dataclass
class OrgProperty(TextListProperty):

    name: str = field(default="org", init=False)


@dataclass
class MemberProperty(UriProperty):

    name: str = field(default="member", init=False)


@dataclass
class RelatedProperty(MultipleValueProperty):

    name: str = field(default="related", init=False)


@dataclass
class CategoriesProperty(TextListProperty):

    name: str = field(default="categories", init=False)


@dataclass
class NoteProperty(TextProperty):

    name: str = field(default="note", init=False)


@dataclass
class ProdidProperty(TextProperty):

    name: str = field(default="prodid", init=False)


@dataclass
class RevProperty(TextProperty):

    name: str = field(default="rev", init=False)

    @classmethod
    def from_node(cls, node: Node) -> RevProperty:  # type: ignore
        name = node.getName()
        if name != cls.name:
            raise ValueError(f"invalid property name: {name}")

        timestamp = node.getTagData("timestamp")
        if not timestamp:
            raise ValueError("no value found")

        parameters = get_parameters(node)

        return cls(timestamp, parameters)

    def to_node(self) -> Node:
        node = Node(self.name)
        if self.parameters:
            node.addChild(node=make_parameters(self.parameters))
        node.addChild("timestamp", payload=self.value)
        return node


@dataclass
class SoundProperty(UriProperty):

    name: str = field(default="sound", init=False)


@dataclass
class UidProperty(UriProperty):

    name: str = field(default="uid", init=False)


@dataclass
class ClientpidmapProperty:

    name: str = field(default="clientpidmap", init=False)
    sourceid: int
    uri: str
    parameters: Parameters = field(default_factory=Parameters)

    @classmethod
    def from_node(cls, node: Node) -> ClientpidmapProperty:
        name = node.getName()
        if name != cls.name:
            raise ValueError(f"invalid property name: {name}")

        sourceid = node.getTagData("sourceid")
        if not sourceid:
            raise ValueError("no value found")

        uri = node.getTagData("uri")
        if not uri:
            raise ValueError("no value found")

        parameters = get_parameters(node)

        return cls(sourceid, uri, parameters)

    def to_node(self) -> Node:
        node = Node(self.name)
        if self.parameters:
            node.addChild(node=make_parameters(self.parameters))
        node.addChild("sourceid", payload=self.sourceid)
        node.addChild("uri", payload=self.uri)
        return node

    @property
    def is_empty(self) -> bool:
        return not self.uri

    def copy(self) -> ClientpidmapProperty:
        return self.__class__(
            sourceid=self.sourceid, uri=self.uri, parameters=self.parameters.copy()
        )


@dataclass
class UrlProperty(UriProperty):

    name: str = field(default="url", init=False)


@dataclass
class KeyProperty(MultipleValueProperty):

    name: str = field(default="key", init=False)


@dataclass
class FBurlProperty(UriProperty):

    name: str = field(default="fburl", init=False)


@dataclass
class CaladruriProperty(UriProperty):

    name: str = field(default="caladruri", init=False)


@dataclass
class CaluriProperty(UriProperty):

    name: str = field(default="calurl", init=False)


PROPERTY_CLASSES = {
    "source": SourceProperty,
    "kind": KindProperty,
    "fn": FnProperty,
    "n": NProperty,
    "nickname": NicknameProperty,
    "photo": PhotoProperty,
    "bday": BDayProperty,
    "anniversary": AnniversaryProperty,
    "gender": GenderProperty,
    "adr": AdrProperty,
    "tel": TelProperty,
    "email": EmailProperty,
    "impp": ImppProperty,
    "lang": LangProperty,
    "tz": TzProperty,
    "geo": GeoProperty,
    "title": TitleProperty,
    "role": RoleProperty,
    "logo": LogoProperty,
    "org": OrgProperty,
    "member": MemberProperty,
    "related": RelatedProperty,
    "categories": CategoriesProperty,
    "note": NoteProperty,
    "prodid": ProdidProperty,
    "rev": RevProperty,
    "sound": SoundProperty,
    "uid": UidProperty,
    "clientpidmap": ClientpidmapProperty,
    "url": UrlProperty,
    "key": KeyProperty,
    "fburl": FBurlProperty,
    "caladruri": CaladruriProperty,
    "caluri": CaluriProperty,
}

PropertyT = (
    SourceProperty
    | KindProperty
    | FnProperty
    | NProperty
    | NicknameProperty
    | PhotoProperty
    | BDayProperty
    | AnniversaryProperty
    | GenderProperty
    | AdrProperty
    | TelProperty
    | EmailProperty
    | ImppProperty
    | LangProperty
    | TzProperty
    | GeoProperty
    | TitleProperty
    | RoleProperty
    | LogoProperty
    | OrgProperty
    | MemberProperty
    | RelatedProperty
    | CategoriesProperty
    | NoteProperty
    | ProdidProperty
    | RevProperty
    | SoundProperty
    | UidProperty
    | ClientpidmapProperty
    | UrlProperty
    | KeyProperty
    | FBurlProperty
    | CaladruriProperty
    | CaluriProperty
)


def get_property_from_name(name: str, node: Node) -> PropertyT | None:
    property_class = PROPERTY_CLASSES.get(name)
    if property_class is None:
        return None

    try:
        return property_class.from_node(node)
    except Exception as error:
        log.warning("invalid vcard property: %s %s", error, node)
        return None


VCardPropertiesT = list[tuple[str | None, PropertyT | list[PropertyT]]]
NotSupportedNodesT = dict[str | None, list[Node]]


class VCard:
    def __init__(
        self,
        properties: VCardPropertiesT | None = None,
        not_supported_nodes: NotSupportedNodesT | None = None,
    ) -> None:
        if properties is None:
            properties = []
        if not_supported_nodes is None:
            not_supported_nodes = defaultdict(list)

        self._properties = properties
        self._no_supported_nodes = not_supported_nodes

    @classmethod
    def from_node(cls, node: Node) -> VCard:
        properties: VCardPropertiesT = []
        not_supported_nodes: NotSupportedNodesT = defaultdict(list)

        for child in node.getChildren():
            if isinstance(child, str):
                continue
            child_name = child.getName()

            if child_name == "group":
                group_name = child.getAttr("name")
                if not group_name:
                    continue

                group_properties: list[PropertyT] = []
                for group_child in child.getChildren():
                    if isinstance(group_child, str):
                        continue

                    group_child_name = group_child.getName()
                    property_ = get_property_from_name(group_child_name, group_child)
                    if property_ is None:
                        not_supported_nodes[group_name].append(group_child)
                        continue
                    group_properties.append(property_)

                properties.append((group_name, group_properties))

            else:

                property_ = get_property_from_name(child_name, child)
                if property_ is None:
                    not_supported_nodes[None].append(child)
                    continue
                properties.append((None, property_))

        return cls(properties, not_supported_nodes)

    def to_node(self) -> Node:
        vcard = Node(f"{Namespace.VCARD4} vcard")
        for group, props in self._properties:
            if group is None:
                assert not isinstance(props, list)
                vcard.addChild(node=props.to_node())
            else:
                group_node = Node(group)
                assert isinstance(props, list)
                for prop in props:
                    group_node.addChild(node=prop.to_node())

                for node in self._no_supported_nodes.get(group, []):
                    group_node.addChild(node=node)

                vcard.addChild(node=group_node)

        for node in self._no_supported_nodes.get(None, []):
            vcard.addChild(node=node)

        return vcard

    def get_properties(self) -> list[PropertyT]:
        properties: list[PropertyT] = []
        for group, props in self._properties:
            if group is None:
                assert not isinstance(props, list)
                properties.append(props)
            else:
                assert isinstance(props, list)
                properties.extend(props)
        return properties

    def add_property(self, name: str, *args: Any, **kwargs: Any):
        prop = PROPERTY_CLASSES.get(name)(*args, **kwargs)
        self._properties.append((None, prop))
        return prop

    def remove_property(self, prop: PropertyT) -> None:
        for _group, props in list(self._properties):
            if isinstance(props, list):
                if prop in props:
                    props.remove(prop)
                    return

            elif prop is props:
                self._properties.remove((None, props))
                return

        raise ValueError("prop not found in vcard")

    def copy(self) -> VCard:
        properties: VCardPropertiesT = []
        for group_name, props in self._properties:
            if group_name is None:
                assert not isinstance(props, list)
                properties.append((None, props.copy()))
            else:
                assert isinstance(props, list)
                group_properties = [prop.copy() for prop in props]
                properties.append((group_name, group_properties))
        return self.__class__(properties=properties)


class VCard4(BaseModule):

    _depends = {
        "publish": "PubSub",
        "request_items": "PubSub",
    }

    def __init__(self, client: Client) -> None:
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = []

    @iq_request_task
    def request_vcard(self, jid: JID | None = None):
        task = yield

        items = yield self.request_items(Namespace.VCARD4_PUBSUB, jid=jid, max_items=1)

        raise_if_error(items)

        if not items:
            yield task.set_result(None)

        yield _get_vcard(items[0])

    @iq_request_task
    def set_vcard(self, vcard: VCard, public: bool = False):
        task = yield

        access_model = "open" if public else "presence"

        options = {
            "pubsub#persist_items": "true",
            "pubsub#access_model": access_model,
        }

        result = yield self.publish(
            Namespace.VCARD4_PUBSUB,
            vcard.to_node(),
            id_="current",
            options=options,
            force_node_options=True,
        )

        yield finalize(task, result)


def _get_vcard(item: Node) -> VCard:
    vcard = item.getTag("vcard", namespace=Namespace.VCARD4)
    if vcard is None:
        raise MalformedStanzaError("vcard node missing", item)

    try:
        vcard = VCard.from_node(vcard)
    except Exception as error:
        raise MalformedStanzaError("invalid vcard: %s" % error, item)

    return vcard



# Copyright (C) 2020 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from typing import Any

import logging

from gi.repository import Gio

from nbxmpp.const import ConnectionType
from nbxmpp.const import TCPState
from nbxmpp.structs import ServerAddress
from nbxmpp.util import LogAdapter
from nbxmpp.util import min_version
from nbxmpp.util import Observable

log = logging.getLogger("nbxmpp.connection")


class Connection(Observable):
    """
    Base Connection Class

    Signals:

        data-sent
        data-received
        bad-certificate
        connection-failed
        disconnected
    """

    def __init__(
        self,
        log_context: str | None,
        address: ServerAddress,
        accepted_certificates: list[Gio.TlsCertificate],
        ignore_tls_errors: bool,
        ignored_tls_errors: set[Gio.TlsCertificateFlags],
        client_cert: Any,
    ) -> None:

        self._log = LogAdapter(log, {"context": log_context})

        Observable.__init__(self, self._log)

        self._client_cert = client_cert
        self._address = address
        self._local_address: Gio.SocketAddress | None = None
        self._remote_address: str | None = None

        self._state = TCPState.DISCONNECTED
        self._tls_con: Gio.TlsConnection | None = None

        self._peer_certificate: Gio.TlsCertificate | None = None
        self._peer_certificate_errors: set[Gio.TlsCertificateFlags] | None = None
        self._accepted_certificates = accepted_certificates
        self._ignore_tls_errors = ignore_tls_errors
        self._ignored_tls_errors = ignored_tls_errors

    @property
    def tls_version(self) -> int | None:
        if self._tls_con is None:
            return None

        if not min_version("GLib", "2.69.0"):
            return None

        return self._tls_con.get_protocol_version()

    @property
    def ciphersuite(self) -> str | None:
        if self._tls_con is None:
            return None

        if not min_version("GLib", "2.69.0"):
            return None

        return self._tls_con.get_ciphersuite_name()

    def get_channel_binding_data(
        self, type_: Gio.TlsChannelBindingType
    ) -> bytes | None:
        assert self._tls_con is not None

        try:
            success, data = self._tls_con.get_channel_binding_data(type_)
        except Exception as error:
            self._log.warning("Unable to get channel binding data: %s", error)
            return None

        if not success:
            return None
        return data

    @property
    def local_address(self) -> Gio.SocketAddress | None:
        return self._local_address

    @property
    def remote_address(self) -> str | None:
        return self._remote_address

    @property
    def peer_certificate(
        self,
    ) -> tuple[Gio.TlsCertificate | None, set[Gio.TlsCertificateFlags] | None]:
        return (self._peer_certificate, self._peer_certificate_errors)

    @property
    def connection_type(self) -> ConnectionType:
        assert self._address is not None
        return self._address.type

    @property
    def state(self) -> TCPState:
        return self._state

    @state.setter
    def state(self, value: TCPState) -> None:
        self._log.info("Set Connection State: %s", value)
        self._state = value

    def _accept_certificate(self) -> bool:
        if not self._peer_certificate_errors:
            return True

        self._log.info(
            "Found TLS certificate errors: %s", self._peer_certificate_errors
        )

        if self._ignore_tls_errors:
            self._log.warning("Ignore all errors")
            return True

        if self._ignored_tls_errors:
            self._log.warning(
                "Ignore TLS certificate errors: %s", self._ignored_tls_errors
            )
            self._peer_certificate_errors -= self._ignored_tls_errors

        if Gio.TlsCertificateFlags.UNKNOWN_CA in self._peer_certificate_errors:
            for accepted_certificate in self._accepted_certificates:
                assert self._peer_certificate is not None
                if self._peer_certificate.is_same(accepted_certificate):
                    self._peer_certificate_errors.discard(
                        Gio.TlsCertificateFlags.UNKNOWN_CA
                    )
                    break

        return bool(not self._peer_certificate_errors)

    def disconnect(self) -> None:
        raise NotImplementedError

    def connect(self) -> None:
        raise NotImplementedError

    def send(self, stanza: Any, now: bool = False) -> None:
        raise NotImplementedError

    def _log_stanza(self, data: str, received: bool = True) -> None:
        direction = "RECEIVED" if received else "SENT"
        message = "::::: DATA %s ::::\n\n%s\n"
        self._log.info(message, direction, data)

    def start_tls_negotiation(self) -> None:
        raise NotImplementedError

    def shutdown_output(self) -> None:
        raise NotImplementedError

    def shutdown_input(self) -> None:
        raise NotImplementedError

    def destroy(self) -> None:
        self.remove_subscriptions()
        self._peer_certificate = None
        self._client_cert = None
        self._address = None
        self._tls_con = None



# Copyright (C) 2020 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from typing import Any
from typing import Literal
from typing import overload
from typing import TYPE_CHECKING

import logging
from collections.abc import Callable
from collections.abc import Iterator
from collections.abc import Sequence

from gi.repository import Gio
from gi.repository import GLib

from nbxmpp.addresses import NoMoreAddresses
from nbxmpp.addresses import ServerAddress
from nbxmpp.addresses import ServerAddresses
from nbxmpp.const import ConnectionProtocol
from nbxmpp.const import ConnectionType
from nbxmpp.const import Mode
from nbxmpp.const import StreamError
from nbxmpp.const import StreamState
from nbxmpp.dispatcher import StanzaDispatcher
from nbxmpp.errors import CancelledError
from nbxmpp.errors import StanzaError
from nbxmpp.errors import TimeoutStanzaError
from nbxmpp.http import HTTPSession
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import BindRequest
from nbxmpp.protocol import Features
from nbxmpp.protocol import isResultNode
from nbxmpp.protocol import JID
from nbxmpp.protocol import Protocol
from nbxmpp.protocol import SessionRequest
from nbxmpp.protocol import StanzaMalformed
from nbxmpp.protocol import TLSRequest
from nbxmpp.protocol import WebsocketCloseHeader
from nbxmpp.sasl import SASL
from nbxmpp.simplexml import Node
from nbxmpp.smacks import Smacks
from nbxmpp.structs import ProxyData
from nbxmpp.task import Task
from nbxmpp.tcp import TCPConnection
from nbxmpp.types import CustomHostT
from nbxmpp.util import generate_id
from nbxmpp.util import get_stream_header
from nbxmpp.util import LogAdapter
from nbxmpp.util import Observable
from nbxmpp.util import validate_stream_header
from nbxmpp.websocket import WebsocketConnection

if TYPE_CHECKING:
    from nbxmpp.connection import Connection
    from nbxmpp.dispatcher import NBXMPPModuleNameT
    from nbxmpp.dispatcher import NBXMPPModuleT
    from nbxmpp.modules.activity import Activity
    from nbxmpp.modules.adhoc import AdHoc
    from nbxmpp.modules.annotations import Annotations
    from nbxmpp.modules.attention import Attention
    from nbxmpp.modules.blocking import Blocking
    from nbxmpp.modules.bookmarks.native_bookmarks import NativeBookmarks
    from nbxmpp.modules.bookmarks.pep_bookmarks import PEPBookmarks
    from nbxmpp.modules.bookmarks.private_bookmarks import PrivateBookmarks
    from nbxmpp.modules.captcha import Captcha
    from nbxmpp.modules.chat_markers import ChatMarkers
    from nbxmpp.modules.chatstates import Chatstates
    from nbxmpp.modules.correction import Correction
    from nbxmpp.modules.delay import Delay
    from nbxmpp.modules.delimiter import Delimiter
    from nbxmpp.modules.discovery import Discovery
    from nbxmpp.modules.eme import EME
    from nbxmpp.modules.entity_caps import EntityCaps
    from nbxmpp.modules.entity_time import EntityTime
    from nbxmpp.modules.http_auth import HTTPAuth
    from nbxmpp.modules.http_upload import HTTPUpload
    from nbxmpp.modules.ibb import IBB
    from nbxmpp.modules.idle import Idle
    from nbxmpp.modules.iq import BaseIq
    from nbxmpp.modules.last_activity import LastActivity
    from nbxmpp.modules.location import Location
    from nbxmpp.modules.mam import MAM
    from nbxmpp.modules.mds import MDS
    from nbxmpp.modules.message import BaseMessage
    from nbxmpp.modules.mood import Mood
    from nbxmpp.modules.muc import MUC
    from nbxmpp.modules.muc.hats import Hats
    from nbxmpp.modules.muc.moderation import Moderation
    from nbxmpp.modules.muclumbus import Muclumbus
    from nbxmpp.modules.nickname import Nickname
    from nbxmpp.modules.omemo import OMEMO
    from nbxmpp.modules.oob import OOB
    from nbxmpp.modules.openpgp import OpenPGP
    from nbxmpp.modules.pgplegacy import PGPLegacy
    from nbxmpp.modules.ping import Ping
    from nbxmpp.modules.presence import BasePresence
    from nbxmpp.modules.pubsub import PubSub
    from nbxmpp.modules.reactions import Reactions
    from nbxmpp.modules.receipts import Receipts
    from nbxmpp.modules.register import Register
    from nbxmpp.modules.replies import Replies
    from nbxmpp.modules.retraction import Retraction
    from nbxmpp.modules.roster import Roster
    from nbxmpp.modules.security_labels import SecurityLabels
    from nbxmpp.modules.software_version import SoftwareVersion
    from nbxmpp.modules.tune import Tune
    from nbxmpp.modules.user_avatar import UserAvatar
    from nbxmpp.modules.vcard4 import VCard4
    from nbxmpp.modules.vcard_avatar import VCardAvatar
    from nbxmpp.modules.vcard_temp import VCardTemp


log = logging.getLogger("nbxmpp.stream")


class Client(Observable):
    def __init__(self, log_context: str | None = None) -> None:
        """
        Signals:
            resume-failed
            resume-successful
            login-successful
            anonymous-supported
            disconnected
            connected
            connection-failed
            stanza-sent
            stanza-received
        """

        self._log_context = log_context
        if log_context is None:
            self._log_context = str(id(self))

        self._log = LogAdapter(log, {"context": self._log_context})

        Observable.__init__(self, self._log)

        self._jid: JID | None = None
        self._lang: str = "en"
        self._domain: str | None = None
        self._username: str | None = None
        self._resource: str | None = None

        self._custom_host: CustomHostT | None = None

        self._addresses: ServerAddresses | None = None
        self._current_address: ServerAddress | None = None
        self._address_generator: Iterator[ServerAddress] | None = None

        self._client_cert: Any = None
        self._client_cert_pass: str | None = None
        self._proxy: ProxyData | None = None
        self._http_session = HTTPSession()

        self._allowed_con_types: list[ConnectionType] | None = None
        self._allowed_protocols: list[ConnectionProtocol] | None = None
        self._allowed_mechs: set[str] | None = None

        self._sm_disabled = False

        self._stream_id: str | None = None
        self._stream_secure = False
        self._stream_authenticated = False
        self._stream_features: Features | None = None
        self._session_required = False
        self._connect_successful = False
        self._stream_close_initiated = False
        self._ping_task: Task | None = None
        self._error: tuple[StreamError | None, str | None, str | None] = (
            None,
            None,
            None,
        )

        self._ignored_tls_errors: set[Gio.TlsCertificateFlags] = set()
        self._ignore_tls_errors = False
        self._accepted_certificates: list[Gio.TlsCertificate] = []
        self._peer_certificate: Gio.TlsCertificate | None = None
        self._peer_certificate_errors: set[Gio.TlsCertificateFlags] | None = None

        self._con: Connection | None = None
        self._local_address: Gio.SocketAddress | None = None
        self._remote_address: str | None = None
        self._mode: Mode = Mode.CLIENT

        self._fallback_ns: set[str] = set()

        self._ping_source_id: int | None = None
        self._tasks: list[Task] = []

        self._dispatcher = StanzaDispatcher(self)
        self._dispatcher.subscribe("before-dispatch", self._on_before_dispatch)
        self._dispatcher.subscribe("parsing-error", self._on_parsing_error)
        self._dispatcher.subscribe("stream-end", self._on_stream_end)

        self._smacks = Smacks(self)
        self._sasl = SASL(self)

        self._state: StreamState = StreamState.DISCONNECTED

    def add_task(self, task: Task) -> None:
        self._tasks.append(task)

    def remove_task(self, task: Task, _context: Any) -> None:
        try:
            self._tasks.remove(task)
        except Exception:
            pass

    @property
    def log_context(self) -> str:
        assert self._log_context is not None
        return self._log_context

    @property
    def features(self) -> Features | None:
        return self._stream_features

    @property
    def resumeable(self) -> bool:
        assert self._smacks is not None
        return self._smacks.resumeable

    @property
    def sm_supported(self) -> bool:
        assert self._smacks is not None
        return self._smacks.sm_supported

    @property
    def lang(self) -> str:
        return self._lang

    def set_lang(self, lang: str) -> None:
        self._lang = lang

    @property
    def username(self) -> str | None:
        return self._username

    @property
    def domain(self) -> str | None:
        return self._domain

    @property
    def resource(self) -> str | None:
        return self._resource

    def set_username(self, username: str | None) -> None:
        self._username = username

    def set_domain(self, domain: str | None) -> None:
        self._domain = domain

    def set_resource(self, resource: str) -> None:
        self._resource = resource

    def set_mode(self, mode: Mode) -> None:
        self._mode = mode

    @property
    def custom_host(self) -> CustomHostT | None:
        return self._custom_host

    def set_custom_host(
        self, host_or_uri: str, protocol: ConnectionProtocol, type_: ConnectionType
    ) -> None:
        if self._domain is None:
            raise ValueError("Call set_domain() first before set_custom_host()")
        self._custom_host = (host_or_uri, protocol, type_)

    def set_accepted_certificates(self, certificates: list[Gio.TlsCertificate]) -> None:
        self._accepted_certificates = certificates

    @property
    def ignored_tls_errors(self) -> set[Gio.TlsCertificateFlags]:
        return set(self._ignored_tls_errors)

    def set_ignored_tls_errors(
        self, errors: set[Gio.TlsCertificateFlags] | None
    ) -> None:
        if errors is None:
            errors = set()
        self._ignored_tls_errors = errors

    @property
    def ignore_tls_errors(self) -> bool:
        return self._ignore_tls_errors

    @property
    def tls_version(self) -> int | None:
        assert self._con is not None
        return self._con.tls_version

    @property
    def ciphersuite(self) -> str | None:
        assert self._con is not None
        return self._con.ciphersuite

    def get_channel_binding_data(
        self, type_: Gio.TlsChannelBindingType
    ) -> bytes | None:
        assert self._con is not None
        return self._con.get_channel_binding_data(type_)

    def set_ignore_tls_errors(self, ignore: bool) -> None:
        self._ignore_tls_errors = ignore

    def set_password(self, password: str | None) -> None:
        assert self._sasl is not None
        self._sasl.set_password(password)

    @property
    def password(self) -> str | None:
        assert self._sasl is not None
        return self._sasl.password

    @property
    def peer_certificate(
        self,
    ) -> tuple[Gio.TlsCertificate | None, set[Gio.TlsCertificateFlags] | None]:
        if self._con is not None:
            return self._con.peer_certificate
        return self._peer_certificate, self._peer_certificate_errors

    @property
    def current_address(self) -> ServerAddress | None:
        return self._current_address

    @property
    def current_connection_type(self) -> ConnectionType:
        assert self._current_address is not None
        return self._current_address.type

    @property
    def is_websocket(self) -> bool:
        assert self._current_address is not None
        return self._current_address.protocol == ConnectionProtocol.WEBSOCKET

    @property
    def stream_id(self) -> str | None:
        return self._stream_id

    @property
    def is_stream_secure(self) -> bool:
        direct_tls = self.current_connection_type == ConnectionType.DIRECT_TLS
        return self._stream_secure or direct_tls

    @property
    def is_stream_authenticated(self) -> bool:
        return self._stream_authenticated

    @property
    def state(self) -> StreamState:
        return self._state

    @state.setter
    def state(self, value: StreamState) -> None:
        self._state = value
        self._log.info("Set state: %s", value)

    def set_state(self, state: StreamState) -> None:
        self.state = state
        self._xmpp_state_machine()

    @property
    def local_address(self) -> Gio.SocketAddress | None:
        return self._local_address

    @property
    def remote_address(self) -> str | None:
        return self._remote_address

    @property
    def connection_types(self) -> list[ConnectionType]:
        if self._custom_host is not None:
            return [self._custom_host[2]]
        return list(
            self._allowed_con_types
            or [ConnectionType.DIRECT_TLS, ConnectionType.START_TLS]
        )

    def set_connection_types(self, con_types: list[ConnectionType]) -> None:
        self._allowed_con_types = con_types

    @property
    def mechs(self) -> set[str]:
        return set(
            self._allowed_mechs
            or {"SCRAM-SHA-512", "SCRAM-SHA-256", "SCRAM-SHA-1", "PLAIN"}
        )

    def set_mechs(self, mechs: set[str]) -> None:
        self._allowed_mechs = mechs

    @property
    def protocols(self) -> list[ConnectionProtocol]:
        if self._custom_host is not None:
            return [self._custom_host[1]]
        return list(
            self._allowed_protocols
            or [ConnectionProtocol.TCP, ConnectionProtocol.WEBSOCKET]
        )

    def set_protocols(self, protocols: list[ConnectionProtocol]) -> None:
        self._allowed_protocols = protocols

    def set_sm_disabled(self, value: bool) -> None:
        self._sm_disabled = value

    @property
    def sm_disabled(self) -> bool:
        return self._sm_disabled

    @property
    def client_cert(self) -> tuple[Any, str | None]:
        return self._client_cert, self._client_cert_pass

    def set_client_cert(self, client_cert: Any, client_cert_pass: str) -> None:
        self._client_cert = client_cert
        self._client_cert_pass = client_cert_pass

    def set_proxy(self, proxy: ProxyData | None) -> None:
        self._proxy = proxy

    @property
    def proxy(self) -> ProxyData | None:
        return self._proxy

    def set_http_session(self, session: HTTPSession) -> None:
        self._http_session = session

    @property
    def http_session(self) -> HTTPSession:
        return self._http_session

    def get_bound_jid(self) -> JID | None:
        return self._jid

    def _set_bound_jid(self, jid: str) -> None:
        self._jid = JID.from_string(jid)

    def set_supported_fallback_ns(self, ns: Sequence[str]) -> None:
        self._fallback_ns = set(ns)

    def get_supported_fallback_ns(self) -> set[str]:
        return set(self._fallback_ns)

    @property
    def has_error(self) -> bool:
        return self._error[0] is not None

    def get_error(self) -> tuple[StreamError | None, str | None, str | None]:
        return self._error

    def _reset_error(self) -> None:
        self._error = None, None, None

    def _set_error(
        self, domain: StreamError, error: str, text: str | None = None
    ) -> None:
        self._log.info("Set error: %s, %s, %s", domain, error, text)
        self._error = domain, error, text

    def _connect(self) -> None:
        if self._state not in (StreamState.DISCONNECTED, StreamState.RESOLVED):
            self._log.error("Stream can't connect, stream state: %s", self._state)
            return

        self.state = StreamState.CONNECTING
        self._peer_certificate = None
        self._peer_certificate_errors = None
        self._reset_error()

        self._con = self._get_connection(
            self._log_context,
            self._current_address,
            self._accepted_certificates,
            self._ignore_tls_errors,
            self._ignored_tls_errors,
            self.client_cert,
        )

        self._con.subscribe("connected", self._on_connected)
        self._con.subscribe("connection-failed", self._on_connection_failed)
        self._con.subscribe("disconnected", self._on_disconnected)
        self._con.subscribe("data-sent", self._on_data_sent)
        self._con.subscribe("data-received", self._on_data_received)
        self._con.subscribe("bad-certificate", self._on_bad_certificate)
        self._con.connect()

    def _get_connection(self, *args: Any) -> WebsocketConnection | TCPConnection:
        if self.is_websocket:
            return WebsocketConnection(*args)
        return TCPConnection(*args)

    def connect(self) -> None:
        if self._state != StreamState.DISCONNECTED:
            self._log.error("Stream can't reconnect, stream state: %s", self._state)
            return

        if self._connect_successful:
            self._log.info("Reconnect")
            self._connect()
            return

        self._log.info("Connect")
        self._reset_error()
        self.state = StreamState.RESOLVE

        self._addresses = ServerAddresses(self._domain)
        self._addresses.set_http_session(self._http_session)
        self._addresses.set_custom_host(self._custom_host)
        self._addresses.set_proxy(self._proxy)
        self._addresses.subscribe("resolved", self._on_addresses_resolved)
        self._addresses.resolve()

    def _on_addresses_resolved(
        self, _addresses: ServerAddresses, _signal_name: str
    ) -> None:
        self._log.info("Domain resolved")
        self._log.info(self._addresses)
        self.state = StreamState.RESOLVED
        assert self._addresses is not None
        self._address_generator = self._addresses.get_next_address(
            self.connection_types, self.protocols
        )

        self._try_next_ip()

    def _try_next_ip(self, *args: Any) -> None:
        try:
            assert self._address_generator is not None
            self._current_address = next(self._address_generator)
        except NoMoreAddresses:
            self._current_address = None
            self.state = StreamState.DISCONNECTED
            assert self._addresses is not None
            self._log.error("Unable to connect to %s", self._addresses.domain)
            self._set_error(
                StreamError.CONNECTION_FAILED,
                "connection-failed",
                "Unable to connect to %s" % self._addresses.domain,
            )
            self.notify("connection-failed")
            return

        self._log.info("Current address: %s", self._current_address)
        self._connect()

    def disconnect(self, immediate: bool = False) -> None:
        if self._state == StreamState.RESOLVE:
            assert self._addresses is not None
            self._addresses.cancel_resolve()
            self.state = StreamState.DISCONNECTED
            return

        if self._state == StreamState.CONNECTING:
            self._disconnect()
            return

        if self._state in (StreamState.DISCONNECTED, StreamState.DISCONNECTING):
            self._log.warning("Stream can't disconnect, stream state: %s", self._state)
            return

        self._disconnect(immediate=immediate)

    def _disconnect(self, immediate: bool = True) -> None:
        self.state = StreamState.DISCONNECTING
        self._remove_ping_timer()
        self._cancel_ping_task()

        if not immediate:
            self._stream_close_initiated = True
            assert self._smacks is not None
            self._smacks.close_session()
            self._end_stream()
            self._con.shutdown_output()
        else:
            self._con.disconnect()

    def send(self, stanza: Protocol, *args: Any, **kwargs: Any) -> str:
        # Alias for backwards compat
        return self.send_stanza(stanza)

    def _on_connected(self, connection: Connection, _signal_name: str) -> None:
        self.set_state(StreamState.CONNECTED)
        self._local_address = connection.local_address
        self._remote_address = connection.remote_address

    def _on_disconnected(self, _connection: Connection, _signal_name: str) -> None:
        self.state = StreamState.DISCONNECTED
        for task in self._tasks:
            task.cancel()
        self._remove_ping_timer()
        self._cancel_ping_task()
        self._reset_stream()
        self.notify("disconnected")

    def _on_connection_failed(self, _connection: Connection, _signal_name: str) -> None:
        self.state = StreamState.DISCONNECTED
        self._reset_stream()
        if not self._connect_successful:
            self._try_next_ip()
        else:
            self._set_error(
                StreamError.CONNECTION_FAILED,
                "connection-failed",
                (
                    "Unable to connect to last "
                    f"successful address: {self._current_address}"
                ),
            )
            self.notify("connection-failed")

    def _disconnect_with_error(
        self, error_domain: StreamError, error: str, text: str | None = None
    ) -> None:
        self._set_error(error_domain, error, text)
        self.disconnect()

    def _on_parsing_error(
        self, _dispatcher: StanzaDispatcher, _signal_name: str, error: str | None
    ) -> None:
        if self._state == StreamState.DISCONNECTING:
            # Don't notify about parsing errors if we already ended the stream
            return
        self._disconnect_with_error(StreamError.PARSING, "parsing-error", error)

    def _on_stream_end(
        self, _dispatcher: StanzaDispatcher, _signal_name: str, error: str | None
    ) -> None:
        if not self.has_error:
            self._set_error(StreamError.STREAM, error or "stream-end")

        self._con.shutdown_input()
        if not self._stream_close_initiated:
            self.state = StreamState.DISCONNECTING
            self._remove_ping_timer()
            self._cancel_ping_task()
            assert self._smacks is not None
            self._smacks.close_session()
            self._end_stream()
            self._con.shutdown_output()

    def _reset_stream(self) -> None:
        self._stream_id = None
        self._stream_secure = False
        self._stream_authenticated = False
        self._stream_features = None
        self._session_required = False
        self._con = None

    def _end_stream(self) -> None:
        if self.is_websocket:
            nonza = WebsocketCloseHeader()
        else:
            nonza = "</stream:stream>"
        self.send_nonza(nonza)

    @overload
    def get_module(self, name: Literal["Activity"]) -> Activity: ...
    @overload
    def get_module(self, name: Literal["AdHoc"]) -> AdHoc: ...
    @overload
    def get_module(self, name: Literal["Annotations"]) -> Annotations: ...
    @overload
    def get_module(self, name: Literal["Attention"]) -> Attention: ...
    @overload
    def get_module(self, name: Literal["Blocking"]) -> Blocking: ...
    @overload
    def get_module(self, name: Literal["NativeBookmarks"]) -> NativeBookmarks: ...
    @overload
    def get_module(self, name: Literal["PEPBookmarks"]) -> PEPBookmarks: ...
    @overload
    def get_module(self, name: Literal["PrivateBookmarks"]) -> PrivateBookmarks: ...
    @overload
    def get_module(self, name: Literal["Captcha"]) -> Captcha: ...
    @overload
    def get_module(self, name: Literal["ChatMarkers"]) -> ChatMarkers: ...
    @overload
    def get_module(self, name: Literal["Chatstates"]) -> Chatstates: ...
    @overload
    def get_module(self, name: Literal["Correction"]) -> Correction: ...
    @overload
    def get_module(self, name: Literal["Delay"]) -> Delay: ...
    @overload
    def get_module(self, name: Literal["Delimiter"]) -> Delimiter: ...
    @overload
    def get_module(self, name: Literal["Discovery"]) -> Discovery: ...
    @overload
    def get_module(self, name: Literal["EME"]) -> EME: ...
    @overload
    def get_module(self, name: Literal["EntityCaps"]) -> EntityCaps: ...
    @overload
    def get_module(self, name: Literal["EntityTime"]) -> EntityTime: ...
    @overload
    def get_module(self, name: Literal["Hats"]) -> Hats: ...
    @overload
    def get_module(self, name: Literal["HTTPAuth"]) -> HTTPAuth: ...
    @overload
    def get_module(self, name: Literal["HTTPUpload"]) -> HTTPUpload: ...
    @overload
    def get_module(self, name: Literal["IBB"]) -> IBB: ...
    @overload
    def get_module(self, name: Literal["Idle"]) -> Idle: ...
    @overload
    def get_module(self, name: Literal["BaseIq"]) -> BaseIq: ...
    @overload
    def get_module(self, name: Literal["LastActivity"]) -> LastActivity: ...
    @overload
    def get_module(self, name: Literal["Location"]) -> Location: ...
    @overload
    def get_module(self, name: Literal["MAM"]) -> MAM: ...
    @overload
    def get_module(self, name: Literal["MDS"]) -> MDS: ...
    @overload
    def get_module(self, name: Literal["BaseMessage"]) -> BaseMessage: ...
    @overload
    def get_module(self, name: Literal["Mood"]) -> Mood: ...
    @overload
    def get_module(self, name: Literal["MUC"]) -> MUC: ...
    @overload
    def get_module(self, name: Literal["Moderation"]) -> Moderation: ...
    @overload
    def get_module(self, name: Literal["Muclumbus"]) -> Muclumbus: ...
    @overload
    def get_module(self, name: Literal["Nickname"]) -> Nickname: ...
    @overload
    def get_module(self, name: Literal["OMEMO"]) -> OMEMO: ...
    @overload
    def get_module(self, name: Literal["OOB"]) -> OOB: ...
    @overload
    def get_module(self, name: Literal["OpenPGP"]) -> OpenPGP: ...
    @overload
    def get_module(self, name: Literal["PGPLegacy"]) -> PGPLegacy: ...
    @overload
    def get_module(self, name: Literal["Ping"]) -> Ping: ...
    @overload
    def get_module(self, name: Literal["BasePresence"]) -> BasePresence: ...
    @overload
    def get_module(self, name: Literal["PubSub"]) -> PubSub: ...
    @overload
    def get_module(self, name: Literal["Reactions"]) -> Reactions: ...
    @overload
    def get_module(self, name: Literal["Receipts"]) -> Receipts: ...
    @overload
    def get_module(self, name: Literal["Register"]) -> Register: ...
    @overload
    def get_module(self, name: Literal["Replies"]) -> Replies: ...
    @overload
    def get_module(self, name: Literal["Retraction"]) -> Retraction: ...
    @overload
    def get_module(self, name: Literal["Roster"]) -> Roster: ...
    @overload
    def get_module(self, name: Literal["SecurityLabels"]) -> SecurityLabels: ...
    @overload
    def get_module(self, name: Literal["SoftwareVersion"]) -> SoftwareVersion: ...
    @overload
    def get_module(self, name: Literal["Tune"]) -> Tune: ...
    @overload
    def get_module(self, name: Literal["UserAvatar"]) -> UserAvatar: ...
    @overload
    def get_module(self, name: Literal["VCard4"]) -> VCard4: ...
    @overload
    def get_module(self, name: Literal["VCardAvatar"]) -> VCardAvatar: ...
    @overload
    def get_module(self, name: Literal["VCardTemp"]) -> VCardTemp: ...

    def get_module(self, name: NBXMPPModuleNameT) -> NBXMPPModuleT:
        assert self._dispatcher is not None
        return self._dispatcher.get_module(name)

    def _on_bad_certificate(self, connection: Connection, _signal_name: str) -> None:
        self._peer_certificate, self._peer_certificate_errors = (
            connection.peer_certificate
        )
        self._set_error(StreamError.BAD_CERTIFICATE, "bad certificate")

    def accept_certificate(self) -> None:
        self._log.info("Certificate accepted")
        assert self._peer_certificate is not None
        self._accepted_certificates.append(self._peer_certificate)
        self._connect()

    def _on_data_sent(
        self, _connection: Connection, _signal_name: str, data: Any
    ) -> None:
        self.notify("stanza-sent", data)

    def _on_before_dispatch(
        self, _dispatcher: StanzaDispatcher, _signal_name: str, data: Any
    ) -> None:
        self.notify("stanza-received", data)

    def _on_data_received(
        self, _connection: Connection, _signal_name: str, data: str
    ) -> None:
        self._dispatcher.process_data(data)
        self._reset_ping_timer()

    def _reset_ping_timer(self) -> None:
        if self.is_websocket:
            return

        if not self._mode.is_client:
            return

        if self.state != StreamState.ACTIVE:
            return

        if self._ping_source_id is not None:
            self._log.info("Remove ping timer")
            GLib.source_remove(self._ping_source_id)
            self._ping_source_id = None

        self._log.info("Start ping timer")
        self._ping_source_id = GLib.timeout_add_seconds(180, self._ping)

    def _remove_ping_timer(self) -> None:
        if self._ping_source_id is None:
            return
        self._log.info("Remove ping timer")
        GLib.source_remove(self._ping_source_id)
        self._ping_source_id = None

    def send_stanza(
        self,
        stanza: Protocol | Any,
        now: bool = False,
        callback: Callable[..., Any] | None = None,
        timeout: int | None = None,
        user_data: Any = None,
    ) -> str:

        if user_data is not None and not isinstance(user_data, dict):
            raise ValueError("arg user_data must be of dict type")

        if not isinstance(stanza, Protocol):
            raise ValueError("Nonzas not allowed, use send_nonza()")

        id_ = stanza.getID()
        if id_ is None:
            id_ = generate_id()
            stanza.setID(id_)

        if callback is not None:
            self._dispatcher.add_callback_for_id(id_, callback, timeout, user_data)
        self._con.send(stanza, now)
        assert self._smacks is not None
        self._smacks.save_in_queue(stanza)
        return id_

    def SendAndCallForResponse(
        self, stanza: Protocol, callback: Callable[..., Any], user_data: Any = None
    ) -> None:
        self.send_stanza(stanza, callback=callback, user_data=user_data)

    def send_nonza(self, nonza: Any, now: bool = False) -> None:
        self._con.send(nonza, now)

    def _xmpp_state_machine(self, stanza: Node | None = None) -> None:
        self._log.info("Execute state machine")
        if stanza is not None:
            if stanza.getName() == "error":
                self._log.info("Stream error")
                # TODO:
                # self._disconnect_with_error(StreamError.SASL,
                #                             stanza.get_condition())
                return

        if self.state == StreamState.CONNECTED:
            self._dispatcher.set_dispatch_callback(self._xmpp_state_machine)
            if (
                self.current_connection_type == ConnectionType.DIRECT_TLS
                and not self.is_websocket
            ):
                self._con.start_tls_negotiation()
                self._stream_secure = True
                self._start_stream()
                return

            self._start_stream()

        elif self.state == StreamState.WAIT_FOR_STREAM_START:
            try:
                self._stream_id = validate_stream_header(
                    stanza, self._domain, self.is_websocket
                )
            except StanzaMalformed as error:
                self._log.error(error)
                self._disconnect_with_error(
                    StreamError.STREAM, "stanza-malformed", "Invalid stream header"
                )
                return

            if (
                self._stream_secure
                or self.current_connection_type == ConnectionType.PLAIN
            ):
                # TLS Negotiation succeeded or we are connected PLAIN
                # We received the stream header and consider this as
                # successfully connected, this means we will not try
                # other connection methods if an error happensafterwards
                self._connect_successful = True

            self.state = StreamState.WAIT_FOR_FEATURES

        elif self.state == StreamState.WAIT_FOR_FEATURES:
            if self._stream_authenticated and self._mode.is_login_test:
                self.notify("login-successful")
                self.disconnect()
                return

            if stanza.getName() != "features":
                self._log.error("Invalid response: %s", stanza)
                self._disconnect_with_error(
                    StreamError.STREAM,
                    "stanza-malformed",
                    "Invalid response, expected features",
                )
                return
            self._on_stream_features(Features(stanza))

        elif self.state == StreamState.WAIT_FOR_TLS_PROCEED:
            if stanza.getNamespace() != Namespace.TLS:
                self._disconnect_with_error(
                    StreamError.TLS,
                    "stanza-malformed",
                    "Invalid namespace for TLS response",
                )
                return

            if stanza.getName() == "failure":
                self._disconnect_with_error(StreamError.TLS, "negotiation-failed")
                return

            if stanza.getName() == "proceed":
                self._con.start_tls_negotiation()
                self._stream_secure = True
                self._start_stream()
                return

            self._log.error("Invalid response")
            self._disconnect_with_error(
                StreamError.TLS, "stanza-malformed", "Invalid TLS response"
            )
            return

        elif self.state == StreamState.PROCEED_WITH_AUTH:
            assert self._sasl is not None
            self._sasl.delegate(stanza)

        elif self.state == StreamState.AUTH_SUCCESSFUL:
            self._stream_authenticated = True
            assert self._sasl is not None
            if self._sasl.is_sasl2():
                self.state = StreamState.WAIT_FOR_FEATURES
            else:
                self._start_stream()

        elif self.state == StreamState.AUTH_FAILED:
            assert self._sasl is not None
            self._disconnect_with_error(StreamError.SASL, *self._sasl.error)

        elif self.state == StreamState.WAIT_FOR_BIND:
            self._on_bind(stanza)

        elif self.state == StreamState.BIND_SUCCESSFUL:
            self._dispatcher.clear_iq_callbacks()
            self._dispatcher.set_dispatch_callback(None)
            assert self._smacks is not None
            self._smacks.send_enable()
            self.state = StreamState.ACTIVE
            self.notify("connected")

        elif self.state == StreamState.WAIT_FOR_SESSION:
            self._on_session(stanza)

        elif self.state == StreamState.WAIT_FOR_RESUMED:
            assert self._smacks is not None
            self._smacks.delegate(stanza)

        elif self.state == StreamState.RESUME_FAILED:
            self.notify("resume-failed")
            self._start_bind()

        elif self.state == StreamState.RESUME_SUCCESSFUL:
            self._dispatcher.set_dispatch_callback(None)
            self.state = StreamState.ACTIVE
            self.notify("resume-successful")

    def _on_stream_features(self, features: Features) -> None:
        if self.is_stream_authenticated:
            self._stream_features = features
            assert self._smacks is not None
            self._smacks.sm_supported = features.has_sm()
            self._session_required = features.session_required()
            if self._smacks.resume_supported:
                self._smacks.resume_request()
                self.state = StreamState.WAIT_FOR_RESUMED
            else:
                self._start_bind()

        elif self.is_stream_secure:
            if self._mode.is_register:
                if features.has_register():
                    self.state = StreamState.ACTIVE
                    self._dispatcher.set_dispatch_callback(None)
                    self.notify("connected")
                else:
                    self._disconnect_with_error(
                        StreamError.REGISTER, "register-not-supported"
                    )
                return

            if self._mode.is_anonymous_test:
                if features.has_anonymous():
                    self.notify("anonymous-supported")
                    self.disconnect()
                else:
                    self._disconnect_with_error(
                        StreamError.SASL, "anonymous-not-supported"
                    )
                return

            self._start_auth(features)

        else:
            tls_supported, required = features.has_starttls()
            if self._current_address.type == ConnectionType.PLAIN:
                if tls_supported and required:
                    self._log.error("Server requires TLS")
                    self._disconnect_with_error(StreamError.TLS, "tls-required")
                    return
                self._start_auth(features)
                return

            if not tls_supported:
                self._log.error("Server does not support TLS")
                self._disconnect_with_error(StreamError.TLS, "tls-not-supported")
                return
            self._start_tls()

    def _start_stream(self) -> None:
        self._log.info("Start stream")
        self._stream_id = None
        self._dispatcher.reset_parser()
        header = get_stream_header(self._domain, self._lang, self.is_websocket)
        self.send_nonza(header)
        self.state = StreamState.WAIT_FOR_STREAM_START

    def _start_tls(self) -> None:
        self.send_nonza(TLSRequest())
        self.state = StreamState.WAIT_FOR_TLS_PROCEED

    def _start_auth(self, features: Features) -> None:
        if not features.has_sasl() and not features.has_sasl_2():
            self._log.error("Server does not support SASL")
            self._disconnect_with_error(StreamError.SASL, "sasl-not-supported")
            return
        self.state = StreamState.PROCEED_WITH_AUTH
        assert self._sasl is not None
        self._sasl.start_auth(features)

    def _start_bind(self) -> None:
        self._log.info("Send bind")
        bind_request = BindRequest(self.resource)
        self.send_stanza(bind_request)
        self.state = StreamState.WAIT_FOR_BIND

    def _on_bind(self, stanza: Protocol) -> None:
        if not isResultNode(stanza):
            self._disconnect_with_error(
                StreamError.BIND, stanza.getError(), stanza.getErrorMsg()
            )
            return

        jid = stanza.getTag("bind").getTagData("jid")
        self._log.info("Successfully bound %s", jid)
        self._set_bound_jid(jid)

        if not self._session_required:
            # Server don't want us to initialize a session
            self._log.info("No session required")
            self.set_state(StreamState.BIND_SUCCESSFUL)
        else:
            session_request = SessionRequest()
            self.send_stanza(session_request)
            self.state = StreamState.WAIT_FOR_SESSION

    def _on_session(self, stanza: Protocol) -> None:
        if isResultNode(stanza):
            self._log.info("Successfully started session")
            self.set_state(StreamState.BIND_SUCCESSFUL)
        else:
            self._log.error("Session open failed")
            self._disconnect_with_error(
                StreamError.SESSION, stanza.getError(), stanza.getErrorMsg()
            )

    def _ping(self) -> None:
        self._ping_source_id = None
        self._ping_task = self.get_module("Ping").ping(
            self.domain, timeout=10, callback=self._on_pong
        )

    def _on_pong(self, task: Task) -> None:
        self._ping_task = None

        try:
            task.finish()
        except TimeoutStanzaError:
            self._log.info("Ping timeout")
            self._disconnect(immediate=True)
            return

        except CancelledError:
            return

        except StanzaError:
            pass

        self._log.info("Pong")

    def _cancel_ping_task(self) -> None:
        if self._ping_task is not None:
            self._ping_task.cancel()

    def register_handler(self, *args: Any, **kwargs: Any) -> None:
        self._dispatcher.register_handler(*args, **kwargs)

    def unregister_handler(self, *args: Any, **kwargs: Any) -> None:
        self._dispatcher.unregister_handler(*args, **kwargs)

    def destroy(self) -> None:
        for task in self._tasks:
            task.cancel()
        self._remove_ping_timer()
        self._smacks = None
        self._sasl = None
        self._dispatcher.cleanup()
        self._dispatcher = None
        self.remove_subscriptions()



# Copyright (C) 2018 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from typing import TYPE_CHECKING

from nbxmpp.modules.base import BaseModule
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Iq
from nbxmpp.protocol import NodeProcessed
from nbxmpp.protocol import Presence
from nbxmpp.structs import DiscoIdentity
from nbxmpp.structs import DiscoInfo
from nbxmpp.structs import EntityCapsData
from nbxmpp.structs import IqProperties
from nbxmpp.structs import PresenceProperties
from nbxmpp.structs import StanzaHandler
from nbxmpp.util import compute_caps_hash

if TYPE_CHECKING:
    from nbxmpp.client import Client


class EntityCaps(BaseModule):
    def __init__(self, client: Client) -> None:
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = [
            StanzaHandler(
                name="presence",
                callback=self._process_entity_caps,
                ns=Namespace.CAPS,
                priority=15,
            ),
            StanzaHandler(
                name="iq",
                callback=self._process_disco_info,
                typ="get",
                ns=Namespace.DISCO_INFO,
                priority=20,
            ),
        ]

        self._identities = []
        self._features = []

        self._uri: str | None = None
        self._node: str | None = None
        self._caps: DiscoInfo | None = None
        self._caps_hash: str | None = None

    def _process_disco_info(
        self, client: Client, stanza: Iq, _properties: IqProperties
    ) -> None:
        if self._caps is None:
            return

        node = stanza.getQuerynode()
        if node is not None:
            if self._node != node:
                return

        iq = stanza.buildReply("result")
        if node is not None:
            iq.setQuerynode(node)

        query = iq.getQuery()
        for identity in self._caps.identities:
            query.addChild(node=identity.get_node())

        for feature in self._caps.features:
            query.addChild("feature", attrs={"var": feature})

        self._log.info("Respond with disco info")
        client.send_stanza(iq)
        raise NodeProcessed

    def _process_entity_caps(
        self, _client: Client, stanza: Presence, properties: PresenceProperties
    ) -> None:
        caps = stanza.getTag("c", namespace=Namespace.CAPS)
        if caps is None:
            return

        hash_algo = caps.getAttr("hash")
        if hash_algo != "sha-1":
            self._log.warning("Unsupported hashing algorithm used: %s", hash_algo)
            self._log.warning(stanza)
            return

        node = caps.getAttr("node")
        if not node:
            self._log.warning("node attribute missing")
            self._log.warning(stanza)
            return

        ver = caps.getAttr("ver")
        if not ver:
            self._log.warning("ver attribute missing")
            self._log.warning(stanza)
            return

        properties.entity_caps = EntityCapsData(hash=hash_algo, node=node, ver=ver)

    @property
    def caps(self) -> EntityCapsData | None:
        if self._caps is None:
            return None

        assert self._uri is not None
        assert self._caps_hash is not None
        return EntityCapsData(hash="sha-1", node=self._uri, ver=self._caps_hash)

    def set_caps(
        self, identities: list[DiscoIdentity], features: list[str], uri: str
    ) -> None:
        self._uri = uri
        self._caps = DiscoInfo(None, identities, features, [])
        self._caps_hash = compute_caps_hash(self._caps, compare=False)
        self._node = "%s#%s" % (uri, self._caps_hash)
